@@ -49,6 +49,7 @@ static const char *about = PACKAGE_STRING " " PACKAGE_URL;
 #include "dkg-tcpip-common.hh"
 #include "dkg-gnunet-common.hh"
 #include "dkg-common.hh"
+#include "dkg-io.hh"
 
 int 					pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2], broadcast_pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
 pid_t 					pid[DKGPG_MAX_N];
@@ -56,7 +57,7 @@ std::vector<std::string>		peers;
 bool					instance_forked = false;
 
 std::string				passphrase, userid, ifilename, ofilename, passwords, hostname, port;
-tmcg_octets_t				keyid, subkeyid, pub, sub, uidsig, subsig, sec, ssb, uid;
+tmcg_openpgp_octets_t			keyid, subkeyid, pub, sub, uidsig, subsig, sec, ssb, uid;
 std::map<size_t, size_t>		idx2dkg, dkg2idx;
 mpz_t					dss_p, dss_q, dss_g, dss_h, dss_x_i, dss_xprime_i, dss_y;
 size_t					dss_n, dss_t, dss_i;
@@ -85,10 +86,11 @@ std::string				armored_message;
 
 
 void print_message
-	(const tmcg_octets_t &msg)
+	(const tmcg_openpgp_octets_t &msg)
 {
 	// print out the decrypted message
-	std::cout << "Decrypted message:" << std::endl;
+	if (opt_verbose)
+		std::cerr << "INFO: decrypted message:" << std::endl;
 	for (size_t i = 0; i < msg.size(); i++)
 		std::cout << msg[i];
 }
@@ -121,7 +123,7 @@ void init_dkg
 			dkg_in << dkg_c_ik[i][k] << std::endl;
 	}
 	if (opt_verbose)
-		std::cout << "GennaroJareckiKrawczykRabinDKG(in, ...)" << std::endl;
+		std::cerr << "INFO: GennaroJareckiKrawczykRabinDKG(in, ...)" << std::endl;
 	dkg = new GennaroJareckiKrawczykRabinDKG(dkg_in);
 	if (!dkg->CheckGroup())
 	{
@@ -294,7 +296,8 @@ bool verify_decryption_share
 		if (!dkg->CheckElement(r_i_out))
 			throw false;
 		// check the NIZK argument for sanity
-		if ((mpz_cmpabs(r_out, dkg->q) >= 0) || (mpz_sizeinbase(c_out, 2L) > 256)) // check the size of r and c (NOTE: output size of mpz_shash is fixed)
+		size_t c_len = mpz_shash_len() * 8; // (NOTE: output size of mpz_shash is fixed)
+		if ((mpz_cmpabs(r_out, dkg->q) >= 0) || (mpz_sizeinbase(c_out, 2L) > c_len)) // check the size of r and c
 			throw false;
 		// verify proof of knowledge (equality of discrete logarithms), e.g. see [CGS97]
 		mpz_powm(a, nizk_gk, r_out, dkg->p);
@@ -445,10 +448,10 @@ bool combine_decryption_shares
 			parties.resize(dkg->t + 1); // we need exactly $t + 1$ decryption shares
 		if (opt_verbose)
 		{
-			std::cout << "combine_decryption_shares(): Lagrange interpolation with ";
+			std::cerr << "INFO: combine_decryption_shares(): Lagrange interpolation with ";
 			for (std::vector<size_t>::iterator jt = parties.begin(); jt != parties.end(); ++jt)
-				std::cout << "P_" << *jt << " ";
-			std::cout << std::endl;
+				std::cerr << "P_" << *jt << " ";
+			std::cerr << std::endl;
 		}
 
 		// compute $R = \prod_{i\in\Lambda} r_i^\lambda_{i,\Lambda} \bmod p$ where $\lambda_{i, \Lambda} = \prod_{l\in\Lambda\setminus\{i\}\frac{l}{l-i}}$
@@ -504,7 +507,7 @@ bool combine_decryption_shares
 }
 
 bool decrypt_session_key
-	(tmcg_octets_t &out)
+	(tmcg_openpgp_octets_t &out)
 {
 	// decrypt the session key
 	gcry_sexp_t elgkey;
@@ -538,6 +541,12 @@ void run_instance
 	(size_t whoami, const size_t num_xtests)
 {
 	std::string armored_seckey, thispeer = peers[whoami];
+	if (!check_strict_permissions(thispeer + "_dkg-sec.asc"))
+	{
+		std::cerr << "WARNING: weak permissions of private key file detected" << std::endl;
+		if (!set_strict_permissions(thispeer + "_dkg-sec.asc"))
+			exit(-1);
+	}
 	if (!read_key_file(thispeer + "_dkg-sec.asc", armored_seckey))
 		exit(-1);
 	init_mpis();
@@ -566,7 +575,7 @@ void run_instance
 			exit(-1);
 		}
 	}
-	tmcg_octets_t enc;
+	tmcg_openpgp_octets_t enc;
 	bool have_seipd = false;
 	GennaroJareckiKrawczykRabinDKG *dkg = NULL;
 	init_dkg(dkg);
@@ -588,7 +597,7 @@ void run_instance
 			std::string pwd;
 			if (!TMCG_ParseHelper::gs(passwords, '/', pwd))
 			{
-				std::cerr << "D_" << whoami << ": " << "cannot read password for protecting channel to D_" << i << std::endl;
+				std::cerr << "ERROR: D_" << whoami << ": " << "cannot read password for protecting channel to D_" << i << std::endl;
 				release_mpis();
 				done_dkg(dkg);
 				exit(-1);
@@ -596,7 +605,7 @@ void run_instance
 			key << pwd;
 			if (((i + 1) < peers.size()) && !TMCG_ParseHelper::nx(passwords, '/'))
 			{
-				std::cerr << "D_" << whoami << ": " << "cannot skip to next password for protecting channel to D_" << (i + 1) << std::endl;
+				std::cerr << "ERROR: D_" << whoami << ": " << "cannot skip to next password for protecting channel to D_" << (i + 1) << std::endl;
 				release_mpis();
 				done_dkg(dkg);
 				exit(-1);
@@ -632,16 +641,16 @@ void run_instance
 	{
 		mpz_t xtest;
 		mpz_init_set_ui(xtest, i);
-		std::cout << "D_" << whoami << ": xtest = " << xtest << " <-> ";
+		std::cerr << "INFO: D_" << whoami << ": xtest = " << xtest << " <-> ";
 		rbc->Broadcast(xtest);
 		for (size_t ii = 0; ii < peers.size(); ii++)
 		{
 			if (!rbc->DeliverFrom(xtest, ii))
-				std::cout << "<X> ";
+				std::cerr << "<X> ";
 			else
-				std::cout << xtest << " ";
+				std::cerr << xtest << " ";
 		}
-		std::cout << std::endl;
+		std::cerr << std::endl;
 		mpz_clear(xtest);
 	}
 
@@ -693,7 +702,7 @@ void run_instance
 	BarnettSmartVTMF_dlog *vtmf = new BarnettSmartVTMF_dlog(crss, TMCG_DDH_SIZE, TMCG_DLSE_SIZE, false); // without verifiable generation of $g$ due to possible FIPS-CRS
 	if (!vtmf->CheckGroup())
 	{
-		std::cout << "D_" << whoami << ": " << "VTMF: Group G was not correctly generated!" << std::endl;
+		std::cerr << "ERROR: D_" << whoami << ": " << "VTMF: Group G was not correctly generated!" << std::endl;
 		delete vtmf;
 		mpz_clear(crs_p), mpz_clear(crs_q), mpz_clear(crs_g), mpz_clear(crs_k);
 		delete aiou, delete aiou2, delete rbc;
@@ -705,7 +714,7 @@ void run_instance
 	// create and exchange keys in order to bootstrap the $h$-generation for EDCF [JL00]
 	// TODO: replace N-time NIZK by one interactive (distributed) zero-knowledge proof of knowledge
 	if (opt_verbose)
-		std::cout << "INFO: generate h for EDCF by using VTMF key generation protocol" << std::endl;
+		std::cerr << "INFO: generate h for EDCF by using VTMF key generation protocol" << std::endl;
 	mpz_t nizk_c, nizk_r, h_j;
 	mpz_init(nizk_c), mpz_init(nizk_r), mpz_init(h_j);
 	vtmf->KeyGenerationProtocol_GenerateKey();
@@ -719,21 +728,21 @@ void run_instance
 		{
 			if (!rbc->DeliverFrom(h_j, i))
 			{
-				std::cerr << "D_" << whoami << ": WARNING - no VTMF key received from " << i << std::endl;
+				std::cerr << "WARNING: D_" << whoami << ": no VTMF key received from D_" << i << std::endl;
 			}
 			if (!rbc->DeliverFrom(nizk_c, i))
 			{
-				std::cerr << "D_" << whoami << ": WARNING - no NIZK c received from " << i << std::endl;
+				std::cerr << "WARNING: D_" << whoami << ": no NIZK c received from D_" << i << std::endl;
 			}
 			if (!rbc->DeliverFrom(nizk_r, i))
 			{
-				std::cerr << "D_" << whoami << ": WARNING - no NIZK r received from " << i << std::endl;
+				std::cerr << "WARNING: D_" << whoami << ": no NIZK r received from D_" << i << std::endl;
 			}
 			std::stringstream lej;
 			lej << h_j << std::endl << nizk_c << std::endl << nizk_r << std::endl;
 			if (!vtmf->KeyGenerationProtocol_UpdateKey(lej))
 			{
-				std::cerr << "D_" << whoami << ": WARNING - VTMF public key of D_" << i <<
+				std::cerr << "WARNING: D_" << whoami << ": VTMF public key of D_" << i <<
 					" was not correctly generated!" << std::endl;
 			}
 		}
@@ -744,7 +753,7 @@ void run_instance
 	// create an instance of the distributed coin-flip protocol (EDCF)
 	size_t T_EDCF = (peers.size() - 1) / 2; // assume maximum synchronous t-resilience for EDCF
 	if (opt_verbose)
-		std::cout << "JareckiLysyanskayaEDCF(" << peers.size() << ", " << T_EDCF << ", ...)" << std::endl;
+		std::cerr << "INFO: JareckiLysyanskayaEDCF(" << peers.size() << ", " << T_EDCF << ", ...)" << std::endl;
 	JareckiLysyanskayaEDCF *edcf = new JareckiLysyanskayaEDCF(peers.size(), T_EDCF, vtmf->p, vtmf->q, vtmf->g, vtmf->h);
 
 	// initialize
@@ -770,7 +779,7 @@ void run_instance
 
 	// collect other decryption shares
 	if (opt_verbose)
-		std::cout << "INFO: start collecting other decryption shares" << std::endl;
+		std::cerr << "INFO: start collecting other decryption shares" << std::endl;
 	std::vector<size_t> complaints;
 	for (size_t i = 0; i < peers.size(); i++)
 	{
@@ -802,8 +811,7 @@ void run_instance
 			if (std::find(complaints.begin(), complaints.end(), i) == complaints.end())
 			{
 				if (opt_verbose)
-					std::cout << "D_" << whoami << ": good decryption share of P_" << idx_dkg << " received from D_" <<
-						i << std::endl;
+					std::cerr << "INFO: D_" << whoami << ": good decryption share of P_" << idx_dkg << " received from D_" << i << std::endl;
 				if (opt_verbose > 1)
 					std::cerr << err_log.str() << std::endl;
 				// collect only verified decryption shares
@@ -825,9 +833,9 @@ void run_instance
 			std::stringstream err_log;
 			prove_decryption_share_interactive_publiccoin(dkg, r_i, aiou, rbc, edcf, err_log);
 			if (opt_verbose)
-				std::cout << "prove_decryption_share_interactive_publiccoin() finished" << std::endl;
+				std::cerr << "INFO: prove_decryption_share_interactive_publiccoin() finished" << std::endl;
 			if (opt_verbose > 1)
-				std::cout << "D_" << whoami << ": log follows" << std::endl << err_log.str();
+				std::cerr << "INFO: D_" << whoami << ": log follows" << std::endl << err_log.str();
 		}
 	}
 
@@ -847,7 +855,7 @@ void run_instance
 	// at the end: deliver some more rounds for waiting parties
 	time_t synctime = aiounicast::aio_timeout_long;
 	if (opt_verbose)
-		std::cout << "D_" << whoami << ": waiting approximately " << (synctime * (T_RBC + 1)) << " seconds for stalled parties" << std::endl;
+		std::cerr << "INFO: D_" << whoami << ": waiting approximately " << (synctime * (T_RBC + 1)) << " seconds for stalled parties" << std::endl;
 	rbc->Sync(synctime);
 
 	// release EDCF
@@ -862,20 +870,20 @@ void run_instance
 	// release handles (unicast channel)
 	uP_in.clear(), uP_out.clear(), uP_key.clear();
 	if (opt_verbose)
-		std::cout << "D_" << whoami << ": aiou.numRead = " << aiou->numRead <<
+		std::cerr << "INFO: D_" << whoami << ": aiou.numRead = " << aiou->numRead <<
 			" aiou.numWrite = " << aiou->numWrite << std::endl;
 
 	// release handles (broadcast channel)
 	bP_in.clear(), bP_out.clear(), bP_key.clear();
 	if (opt_verbose)
-		std::cout << "D_" << whoami << ": aiou2.numRead = " << aiou2->numRead <<
+		std::cerr << "INFO: D_" << whoami << ": aiou2.numRead = " << aiou2->numRead <<
 			" aiou2.numWrite = " << aiou2->numWrite << std::endl;
 
 	// release asynchronous unicast and broadcast
 	delete aiou, delete aiou2;
 
 	// do remaining decryption work
-	tmcg_octets_t msg, seskey;
+	tmcg_openpgp_octets_t msg, seskey;
 	if (res)
 	{
 		if (!decrypt_session_key(seskey))
@@ -927,7 +935,7 @@ void fork_instance
 	(const size_t whoami)
 {
 	if ((pid[whoami] = fork()) < 0)
-		perror("dkg-decrypt (fork)");
+		perror("ERROR: dkg-decrypt (fork)");
 	else
 	{
 		if (pid[whoami] == 0)
@@ -939,14 +947,14 @@ void fork_instance
 			run_instance(whoami, 0);
 #endif
 			if (opt_verbose)
-				std::cout << "D_" << whoami << ": exit(0)" << std::endl;
+				std::cerr << "INFO: D_" << whoami << ": exit(0)" << std::endl;
 			exit(0);
 			/* END child code: participant D_i */
 		}
 		else
 		{
 			if (opt_verbose)
-				std::cout << "fork() = " << pid[whoami] << std::endl;
+				std::cerr << "INFO: fork() = " << pid[whoami] << std::endl;
 			instance_forked = true;
 		}
 	}
@@ -1185,9 +1193,9 @@ int main
 	}
 	if (opt_verbose)
 	{
-		std::cout << "INFO: canonicalized peer list = " << std::endl;
+		std::cerr << "INFO: canonicalized peer list = " << std::endl;
 		for (size_t i = 0; i < peers.size(); i++)
-			std::cout << peers[i] << std::endl;
+			std::cerr << peers[i] << std::endl;
 	}
 
 	// lock memory
@@ -1215,7 +1223,7 @@ int main
 		return -1;
 	}
 	if (opt_verbose)
-		std::cout << "INFO: using LibTMCG version " << version_libTMCG() << std::endl;
+		std::cerr << "INFO: using LibTMCG version " << version_libTMCG() << std::endl;
 
 	// read message
 	if (opt_ifilename != NULL)
@@ -1233,7 +1241,7 @@ int main
 	}
 	else
 	{
-		std::cout << "Please enter the encrypted message (in ASCII Armor; ^D for EOF): " << std::endl;
+		std::cerr << "Please enter the encrypted message (in ASCII Armor; ^D for EOF): " << std::endl;
 		std::string line;
 		while (std::getline(std::cin, line))
 			armored_message += line + "\r\n";
@@ -1244,12 +1252,18 @@ int main
 	if (nonint)
 	{
 		size_t idx;
-		tmcg_octets_t msg, seskey;
+		tmcg_openpgp_octets_t msg, seskey;
 		std::string dds, armored_seckey, thispeer = peers[0];
 		mpz_t r_i, c, r;
 		std::vector<size_t> interpol_parties;
 		std::vector<mpz_ptr> interpol_shares;
 
+		if (!check_strict_permissions(thispeer + "_dkg-sec.asc"))
+		{
+			std::cerr << "WARNING: weak permissions of private key file detected" << std::endl;
+			if (!set_strict_permissions(thispeer + "_dkg-sec.asc"))
+				return -1;
+		}
 		if (!read_key_file(thispeer + "_dkg-sec.asc", armored_seckey))
 			return -1;
 		init_mpis();
@@ -1274,7 +1288,7 @@ int main
 				return -1;
 			}
 		}
-		tmcg_octets_t enc;
+		tmcg_openpgp_octets_t enc;
 		bool have_seipd = false;
 		GennaroJareckiKrawczykRabinDKG *dkg = NULL;
 		init_dkg(dkg);
@@ -1285,17 +1299,17 @@ int main
 			return -1;
 		}
 		compute_decryption_share(dkg, dds);
-		tmcg_octets_t dds_input;
-		dds_input.push_back((tmcg_byte_t)(mpz_wrandom_ui() % 256)); // bluring the decryption share
-		dds_input.push_back((tmcg_byte_t)(mpz_wrandom_ui() % 256)); // make NSA's spying a bit harder
-		dds_input.push_back((tmcg_byte_t)(mpz_wrandom_ui() % 256));
-		dds_input.push_back((tmcg_byte_t)(mpz_wrandom_ui() % 256));
-		dds_input.push_back((tmcg_byte_t)(mpz_wrandom_ui() % 256));
+		tmcg_openpgp_octets_t dds_input;
+		dds_input.push_back((tmcg_openpgp_byte_t)(mpz_wrandom_ui() % 256)); // bluring the decryption share
+		dds_input.push_back((tmcg_openpgp_byte_t)(mpz_wrandom_ui() % 256)); // make NSA's spying a bit harder
+		dds_input.push_back((tmcg_openpgp_byte_t)(mpz_wrandom_ui() % 256));
+		dds_input.push_back((tmcg_openpgp_byte_t)(mpz_wrandom_ui() % 256));
+		dds_input.push_back((tmcg_openpgp_byte_t)(mpz_wrandom_ui() % 256));
 		for (size_t i = 0; i < dds.length(); i++)
 			dds_input.push_back(dds[i]);
 		std::string dds_radix;
 		CallasDonnerhackeFinneyShawThayerRFC4880::Radix64Encode(dds_input, dds_radix, false);
-		std::cout << "Your decryption share (keep confidential): " << dds_radix << std::endl;
+		std::cerr << "Your decryption share (keep confidential): " << dds_radix << std::endl;
 		mpz_init(r_i), mpz_init(c), mpz_init(r);
 		if (!verify_decryption_share(dkg, dds, idx, r_i, c, r))
 		{
@@ -1307,10 +1321,10 @@ int main
 		mpz_ptr tmp1 = new mpz_t();
 		mpz_init_set(tmp1, r_i);
 		interpol_parties.push_back(dkg->i), interpol_shares.push_back(tmp1);
-		std::cout << "Enter decryption shares (one per line; ^D for EOF) from other parties/devices:" << std::endl;
+		std::cerr << "Enter decryption shares (one per line; ^D for EOF) from other parties/devices:" << std::endl;
 		while (std::getline(std::cin, dds_radix))
 		{
-			tmcg_octets_t dds_output;
+			tmcg_openpgp_octets_t dds_output;
 			dds = "", idx = 0;
 			CallasDonnerhackeFinneyShawThayerRFC4880::Radix64Decode(dds_radix, dds_output);
 			for (size_t i = 5; i < dds_output.size(); i++)
@@ -1325,10 +1339,10 @@ int main
 					interpol_parties.push_back(idx), interpol_shares.push_back(tmp1);
 				}
 				else
-					std::cout << "WARNING: decryption share of P_" << idx << " already stored" << std::endl;
+					std::cerr << "WARNING: decryption share of P_" << idx << " already stored" << std::endl;
 			}
 			else
-				std::cout << "WARNING: verification of decryption share from P_" << idx << " failed" << std::endl;
+				std::cerr << "WARNING: verification of decryption share from P_" << idx << " failed" << std::endl;
 		}
 		bool res = combine_decryption_shares(dkg, interpol_parties, interpol_shares);
 		mpz_clear(r_i), mpz_clear(c), mpz_clear(r);
@@ -1464,16 +1478,16 @@ int main
 	std::cerr << "WARNING: GNUnet CADET is required for the message exchange of this program" << std::endl;
 #endif
 
-	std::cout << "INFO: running local test with " << peers.size() << " participants" << std::endl;
+	std::cerr << "INFO: running local test with " << peers.size() << " participants" << std::endl;
 	// open pipes
 	for (size_t i = 0; i < peers.size(); i++)
 	{
 		for (size_t j = 0; j < peers.size(); j++)
 		{
 			if (pipe(pipefd[i][j]) < 0)
-				perror("dkg-decrypt (pipe)");
+				perror("ERROR: dkg-decrypt (pipe)");
 			if (pipe(broadcast_pipefd[i][j]) < 0)
-				perror("dkg-decrypt (pipe)");
+				perror("ERROR: dkg-decrypt (pipe)");
 		}
 	}
 	
@@ -1489,9 +1503,9 @@ int main
 	{
 		int wstatus = 0;
 		if (opt_verbose)
-			std::cout << "waitpid(" << pid[i] << ")" << std::endl;
+			std::cerr << "INFO: waitpid(" << pid[i] << ")" << std::endl;
 		if (waitpid(pid[i], &wstatus, 0) != pid[i])
-			perror("dkg-decrypt (waitpid)");
+			perror("ERROR: dkg-decrypt (waitpid)");
 		if (!WIFEXITED(wstatus))
 		{
 			std::cerr << "ERROR: protocol instance ";
@@ -1504,16 +1518,16 @@ int main
 		else if (WIFEXITED(wstatus))
 		{
 			if (opt_verbose)
-				std::cout << "INFO: protocol instance " << pid[i] << " terminated with exit status " << WEXITSTATUS(wstatus) << std::endl;
+				std::cerr << "INFO: protocol instance " << pid[i] << " terminated with exit status " << WEXITSTATUS(wstatus) << std::endl;
 			if (WEXITSTATUS(wstatus))
 				ret = -2; // error
 		}
 		for (size_t j = 0; j < peers.size(); j++)
 		{
 			if ((close(pipefd[i][j][0]) < 0) || (close(pipefd[i][j][1]) < 0))
-				perror("dkg-decrypt (close)");
+				perror("ERROR: dkg-decrypt (close)");
 			if ((close(broadcast_pipefd[i][j][0]) < 0) || (close(broadcast_pipefd[i][j][1]) < 0))
-				perror("dkg-decrypt (close)");
+				perror("ERROR: dkg-decrypt (close)");
 		}
 	}
 	
