@@ -1,7 +1,7 @@
 /*******************************************************************************
    This file is part of Distributed Privacy Guard (DKGPG).
 
- Copyright (C) 2017  Heiko Stamer <HeikoStamer@gmx.net>
+ Copyright (C) 2017, 2018  Heiko Stamer <HeikoStamer@gmx.net>
 
    DKGPG is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -49,6 +49,7 @@ static const char *about = PACKAGE_STRING " " PACKAGE_URL;
 #include "dkg-tcpip-common.hh"
 #include "dkg-gnunet-common.hh"
 #include "dkg-common.hh"
+#include "dkg-io.hh"
 
 int 					pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2], broadcast_pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
 pid_t 					pid[DKGPG_MAX_N];
@@ -56,7 +57,7 @@ std::vector<std::string>		peers;
 bool					instance_forked = false;
 
 std::string				passphrase, userid, passwords, hostname, port;
-tmcg_octets_t				keyid, subkeyid, pub, sub, uidsig, subsig, sec, ssb, uid;
+tmcg_openpgp_octets_t			keyid, subkeyid, pub, sub, uidsig, subsig, sec, ssb, uid;
 std::map<size_t, size_t>		idx2dkg, dkg2idx;
 mpz_t					dss_p, dss_q, dss_g, dss_h, dss_x_i, dss_xprime_i, dss_y;
 size_t					dss_n, dss_t, dss_i;
@@ -68,9 +69,12 @@ std::vector<size_t>			dkg_qual;
 std::vector<mpz_ptr>			dkg_v_i;
 std::vector< std::vector<mpz_ptr> >	dkg_c_ik;
 gcry_mpi_t 				dsa_p, dsa_q, dsa_g, dsa_y, dsa_x, elg_p, elg_q, elg_g, elg_y, elg_x;
+gcry_mpi_t				dsa_r, dsa_s, elg_r, elg_s, rsa_n, rsa_e, rsa_md;
 gcry_mpi_t 				gk, myk, sig_r, sig_s;
+gcry_mpi_t				revdsa_r, revdsa_s, revelg_r, revelg_s, revrsa_md;
 
 int 					opt_verbose = 0;
+bool					libgcrypt_secmem = false;
 char					*opt_passwords = NULL;
 char					*opt_hostname = NULL;
 unsigned long int			opt_r = 0, opt_p = 55000, opt_W = 5;
@@ -80,6 +84,12 @@ void run_instance
 {
 	// read and parse the private key
 	std::string armored_seckey, thispeer = peers[whoami];
+	if (!check_strict_permissions(thispeer + "_dkg-sec.asc"))
+	{
+		std::cerr << "WARNING: weak permissions of private key file detected" << std::endl;
+		if (!set_strict_permissions(thispeer + "_dkg-sec.asc"))
+			exit(-1);
+	}
 	if (!read_key_file(thispeer + "_dkg-sec.asc", armored_seckey))
 		exit(-1);
 	init_mpis();
@@ -103,7 +113,7 @@ void run_instance
 #endif
 		if (!parse_private_key(armored_seckey, ckeytime, ekeytime, CAPL))
 		{
-			std::cerr << "R_" << whoami << ": wrong passphrase to unlock private key" << std::endl;
+			std::cerr << "ERROR: R_" << whoami << ": wrong passphrase to unlock private key" << std::endl;
 			release_mpis();
 			exit(-1);
 		}
@@ -120,14 +130,14 @@ void run_instance
 			std::string pwd;
 			if (!TMCG_ParseHelper::gs(passwords, '/', pwd))
 			{
-				std::cerr << "R_" << whoami << ": " << "cannot read password for protecting channel to R_" << i << std::endl;
+				std::cerr << "ERROR: R_" << whoami << ": " << "cannot read password for protecting channel to R_" << i << std::endl;
 				release_mpis();
 				exit(-1);
 			}
 			key << pwd;
 			if (((i + 1) < peers.size()) && !TMCG_ParseHelper::nx(passwords, '/'))
 			{
-				std::cerr << "R_" << whoami << ": " << "cannot skip to next password for protecting channel to R_" << (i + 1) << std::endl;
+				std::cerr << "ERROR: R_" << whoami << ": " << "cannot skip to next password for protecting channel to R_" << (i + 1) << std::endl;
 				release_mpis();
 				exit(-1);
 			}
@@ -161,22 +171,22 @@ void run_instance
 	{
 		mpz_t xtest;
 		mpz_init_set_ui(xtest, i);
-		std::cout << "R_" << whoami << ": xtest = " << xtest << " <-> ";
+		std::cerr << "INFO: R_" << whoami << ": xtest = " << xtest << " <-> ";
 		rbc->Broadcast(xtest);
 		for (size_t ii = 0; ii < peers.size(); ii++)
 		{
 			if (!rbc->DeliverFrom(xtest, ii))
-				std::cout << "<X> ";
+				std::cerr << "<X> ";
 			else
-				std::cout << xtest << " ";
+				std::cerr << xtest << " ";
 		}
-		std::cout << std::endl;
+		std::cerr << std::endl;
 		mpz_clear(xtest);
 	}
 
 	// participants must agree on a common signature creation time (OpenPGP)
 	if (opt_verbose)
-		std::cout << "agree on a signature creation time for OpenPGP" << std::endl;
+		std::cerr << "INFO: agree on a signature creation time for OpenPGP" << std::endl;
 	time_t csigtime = 0;
 	std::vector<time_t> tvs;
 	mpz_t mtv;
@@ -195,7 +205,7 @@ void run_instance
 			}
 			else
 			{
-				std::cerr << "R_" << whoami << ": WARNING - no signature creation time stamp received from " << i << std::endl;
+				std::cerr << "WARNING: R_" << whoami << ": no signature creation time stamp received from " << i << std::endl;
 			}
 		}
 	}
@@ -203,26 +213,26 @@ void run_instance
 	std::sort(tvs.begin(), tvs.end());
 	if (tvs.size() < (peers.size() - T_RBC))
 	{
-		std::cerr << "R_" << whoami << ": not enough timestamps received" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": not enough timestamps received" << std::endl;
 		delete rbc, delete aiou, delete aiou2;
 		release_mpis();
 		exit(-1);
 	}
 	csigtime = tvs[tvs.size()/2]; // use a median value as some kind of gentle agreement
 	if (opt_verbose)
-		std::cout << "R_" << whoami << ": canonicalized signature creation time = " << csigtime << std::endl;
+		std::cerr << "INFO: R_" << whoami << ": canonicalized signature creation time = " << csigtime << std::endl;
 
 	// select hash algorithm for OpenPGP based on |q| (size in bit)
-	tmcg_byte_t hashalgo = 0;
+	tmcg_openpgp_hashalgo_t hashalgo = TMCG_OPENPGP_HASHALGO_UNKNOWN;
 	if (mpz_sizeinbase(dss_q, 2L) == 256)
-		hashalgo = 8; // SHA256 (alg 8)
+		hashalgo = TMCG_OPENPGP_HASHALGO_SHA256; // SHA256 (alg 8)
 	else if (mpz_sizeinbase(dss_q, 2L) == 384)
-		hashalgo = 9; // SHA384 (alg 9)
+		hashalgo = TMCG_OPENPGP_HASHALGO_SHA384; // SHA384 (alg 9)
 	else if (mpz_sizeinbase(dss_q, 2L) == 512)
-		hashalgo = 10; // SHA512 (alg 10)
+		hashalgo = TMCG_OPENPGP_HASHALGO_SHA512; // SHA512 (alg 10)
 	else
 	{
-		std::cerr << "R_" << whoami << ": selecting hash algorithm failed for |q| = " << mpz_sizeinbase(dss_q, 2L) << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": selecting hash algorithm failed for |q| = " << mpz_sizeinbase(dss_q, 2L) << std::endl;
 		delete rbc, delete aiou, delete aiou2;
 		release_mpis();
 		exit(-1);
@@ -270,35 +280,39 @@ void run_instance
 			dss_in << dss_c_ik[i][k] << std::endl;
 	}
 	if (opt_verbose)
-		std::cout << "CanettiGennaroJareckiKrawczykRabinDSS(in, ...)" << std::endl;
+		std::cerr << "INFO: CanettiGennaroJareckiKrawczykRabinDSS(in, ...)" << std::endl;
 	CanettiGennaroJareckiKrawczykRabinDSS *dss = new CanettiGennaroJareckiKrawczykRabinDSS(dss_in);
 	if (!dss->CheckGroup())
 	{
-		std::cerr << "R_" << whoami << ": " << "tDSS domain parameters are not correctly generated!" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": " << "tDSS domain parameters are not correctly generated!" << std::endl;
 		delete dss, delete rbc, delete aiou, delete aiou2;
 		release_mpis();
 		exit(-1);
 	}
 	
 	// reason for revocation
-	tmcg_byte_t revcode = 0x00;
+	tmcg_openpgp_byte_t revcode = 0x00;
 	if (reason < 255)
 		revcode = reason;
 
 	// compute a hash of pub and sub, respectively
-	tmcg_octets_t trailer_pub, pub_hashing, hash_pub, left_pub, trailer_sub, sub_hashing, hash_sub, left_sub;
+	tmcg_openpgp_octets_t trailer_pub, pub_hashing, hash_pub, left_pub, trailer_sub, sub_hashing, hash_sub, left_sub;
 	CallasDonnerhackeFinneyShawThayerRFC4880::PacketSigPrepareRevocationSignature(0x20, hashalgo, csigtime, revcode, "", keyid, trailer_pub);
 	CallasDonnerhackeFinneyShawThayerRFC4880::PacketSigPrepareRevocationSignature(0x28, hashalgo, csigtime, revcode, "", keyid, trailer_sub);
 	for (size_t i = 6; i < pub.size(); i++)
 		pub_hashing.push_back(pub[i]);
 	for (size_t i = 6; i < sub.size(); i++)
 		sub_hashing.push_back(sub[i]);
-	CallasDonnerhackeFinneyShawThayerRFC4880::KeyRevocationHash(pub_hashing, trailer_pub, hashalgo, hash_pub, left_pub);
-	CallasDonnerhackeFinneyShawThayerRFC4880::KeyRevocationHash(pub_hashing, sub_hashing, trailer_sub, hashalgo, hash_sub, left_sub);
+	// RFC 4880 ERRATA:
+	// Primary key revocation signatures (type 0x20) hash only the key being revoked.
+	// Subkey revocation signature (type 0x28) hash first the primary key and then the
+	// subkey being revoked.
+	CallasDonnerhackeFinneyShawThayerRFC4880::KeyHash(pub_hashing, trailer_pub, hashalgo, hash_pub, left_pub);
+	CallasDonnerhackeFinneyShawThayerRFC4880::KeyHash(pub_hashing, sub_hashing, trailer_sub, hashalgo, hash_sub, left_sub);
 
 	// sign the hashes
-	tmcg_octets_t revsig_pub, revsig_sub;
-	tmcg_byte_t buffer[1024];
+	tmcg_openpgp_octets_t revsig_pub, revsig_sub;
+	tmcg_openpgp_byte_t buffer[1024];
 	gcry_mpi_t r, s, h;
 	mpz_t dsa_m, dsa_r, dsa_s;
 	size_t buflen = 0;
@@ -312,7 +326,7 @@ void run_instance
 	ret = gcry_mpi_scan(&h, GCRYMPI_FMT_USG, buffer, buflen, NULL);
 	if (ret)
 	{
-		std::cerr << "R_" << whoami << ": gcry_mpi_scan() failed for h" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": gcry_mpi_scan() failed for h" << std::endl;
 		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		delete dss, delete rbc, delete aiou, delete aiou2;
@@ -321,7 +335,7 @@ void run_instance
 	}
 	if (!mpz_set_gcry_mpi(h, dsa_m))
 	{
-		std::cerr << "R_" << whoami << ": mpz_set_gcry_mpi() failed for dsa_m" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": mpz_set_gcry_mpi() failed for dsa_m" << std::endl;
 		gcry_mpi_release(r), gcry_mpi_release(s), gcry_mpi_release(h);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		delete dss, delete rbc, delete aiou, delete aiou2;
@@ -331,11 +345,11 @@ void run_instance
 	gcry_mpi_release(h);
 	std::stringstream err_log_sign;
 	if (opt_verbose)
-		std::cout << "R_" << whoami << ": dss.Sign() on pub" << std::endl;
+		std::cerr << "INFO: R_" << whoami << ": dss.Sign() on pub" << std::endl;
 	if (!dss->Sign(peers.size(), whoami, dsa_m, dsa_r, dsa_s, idx2dkg, dkg2idx, aiou, rbc, err_log_sign))
 	{
-		std::cerr << "R_" << whoami << ": " << "tDSS Sign() on pub failed" << std::endl;
-		std::cerr << "R_" << whoami << ": log follows " << std::endl << err_log_sign.str();
+		std::cerr << "ERROR: R_" << whoami << ": " << "tDSS Sign() on pub failed" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": log follows " << std::endl << err_log_sign.str();
 		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		delete dss, delete rbc, delete aiou, delete aiou2;
@@ -343,10 +357,10 @@ void run_instance
 		exit(-1);
 	}
 	if (opt_verbose > 1)
-		std::cout << "R_" << whoami << ": log follows " << std::endl << err_log_sign.str();
+		std::cerr << "INFO: R_" << whoami << ": log follows " << std::endl << err_log_sign.str();
 	if (!mpz_get_gcry_mpi(r, dsa_r))
 	{
-		std::cerr << "R_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_r" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_r" << std::endl;
 		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		delete dss, delete rbc, delete aiou, delete aiou2;
@@ -355,7 +369,7 @@ void run_instance
 	}
 	if (!mpz_get_gcry_mpi(s, dsa_s))
 	{
-		std::cerr << "R_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_s" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_s" << std::endl;
 		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		delete dss, delete rbc, delete aiou, delete aiou2;
@@ -370,7 +384,7 @@ void run_instance
 	ret = gcry_mpi_scan(&h, GCRYMPI_FMT_USG, buffer, buflen, NULL);
 	if (ret)
 	{
-		std::cerr << "R_" << whoami << ": gcry_mpi_scan() failed for h" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": gcry_mpi_scan() failed for h" << std::endl;
 		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		delete dss, delete rbc, delete aiou, delete aiou2;
@@ -379,7 +393,7 @@ void run_instance
 	}
 	if (!mpz_set_gcry_mpi(h, dsa_m))
 	{
-		std::cerr << "R_" << whoami << ": mpz_set_gcry_mpi() failed for dsa_m" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": mpz_set_gcry_mpi() failed for dsa_m" << std::endl;
 		gcry_mpi_release(r), gcry_mpi_release(s), gcry_mpi_release(h);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		delete dss, delete rbc, delete aiou, delete aiou2;
@@ -389,11 +403,11 @@ void run_instance
 	gcry_mpi_release(h);
 	std::stringstream err_log_sign2;
 	if (opt_verbose)
-		std::cout << "R_" << whoami << ": dss.Sign() on sub" << std::endl;
+		std::cerr << "INFO: R_" << whoami << ": dss.Sign() on sub" << std::endl;
 	if (!dss->Sign(peers.size(), whoami, dsa_m, dsa_r, dsa_s, idx2dkg, dkg2idx, aiou, rbc, err_log_sign2))
 	{
-		std::cerr << "R_" << whoami << ": " << "tDSS Sign() on sub failed" << std::endl;
-		std::cerr << "R_" << whoami << ": log follows " << std::endl << err_log_sign2.str();
+		std::cerr << "ERROR: R_" << whoami << ": " << "tDSS Sign() on sub failed" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": log follows " << std::endl << err_log_sign2.str();
 		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		delete dss, delete rbc, delete aiou, delete aiou2;
@@ -401,10 +415,10 @@ void run_instance
 		exit(-1);
 	}
 	if (opt_verbose > 1)
-		std::cout << "R_" << whoami << ": log follows " << std::endl << err_log_sign2.str();
+		std::cerr << "INFO: R_" << whoami << ": log follows " << std::endl << err_log_sign2.str();
 	if (!mpz_get_gcry_mpi(r, dsa_r))
 	{
-		std::cerr << "R_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_r" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_r" << std::endl;
 		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		delete dss, delete rbc, delete aiou, delete aiou2;
@@ -413,7 +427,7 @@ void run_instance
 	}
 	if (!mpz_get_gcry_mpi(s, dsa_s))
 	{
-		std::cerr << "R_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_s" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_s" << std::endl;
 		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		delete dss, delete rbc, delete aiou, delete aiou2;
@@ -427,14 +441,14 @@ void run_instance
 	// at the end: deliver some more rounds for still waiting parties
 	time_t synctime = aiounicast::aio_timeout_very_long;
 	if (opt_verbose)
-		std::cout << "R_" << whoami << ": waiting approximately " << (synctime * (T_RBC + 1)) << " seconds for stalled parties" << std::endl;
+		std::cerr << "INFO: R_" << whoami << ": waiting approximately " << (synctime * (T_RBC + 1)) << " seconds for stalled parties" << std::endl;
 	rbc->Sync(synctime);
 
 	// export updated public key in OpenPGP armor format
 	std::stringstream pubfilename;
 	pubfilename << peers[whoami] << "_dkg-pub.asc";
 	std::string armor = "";
-	tmcg_octets_t all, uid;
+	tmcg_openpgp_octets_t all, uid;
 	CallasDonnerhackeFinneyShawThayerRFC4880::PacketUidEncode(userid, uid);
 	all.insert(all.end(), pub.begin(), pub.end());
 	all.insert(all.end(), revsig_pub.begin(), revsig_pub.end());
@@ -443,20 +457,20 @@ void run_instance
 	all.insert(all.end(), sub.begin(), sub.end());
 	all.insert(all.end(), revsig_sub.begin(), revsig_sub.end());
 	all.insert(all.end(), subsig.begin(), subsig.end());
-	CallasDonnerhackeFinneyShawThayerRFC4880::ArmorEncode(6, all, armor);
+	CallasDonnerhackeFinneyShawThayerRFC4880::ArmorEncode(TMCG_OPENPGP_ARMOR_PUBLIC_KEY_BLOCK, all, armor);
 	if (opt_verbose > 1)
 		std::cout << armor << std::endl;
 	std::ofstream pubofs((pubfilename.str()).c_str(), std::ofstream::out);
 	if (!pubofs.good())
 	{
-		std::cerr << "R_" << whoami << ": opening public key file failed" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": opening public key file failed" << std::endl;
 		delete dss, delete rbc, delete aiou, delete aiou2;
 		exit(-1);
 	}
 	pubofs << armor;
 	if (!pubofs.good())
 	{
-		std::cerr << "R_" << whoami << ": writing public key file failed" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": writing public key file failed" << std::endl;
 		delete dss, delete rbc, delete aiou, delete aiou2;
 		exit(-1);
 	}
@@ -471,13 +485,13 @@ void run_instance
 	// release handles (unicast channel)
 	uP_in.clear(), uP_out.clear(), uP_key.clear();
 	if (opt_verbose)
-		std::cout << "R_" << whoami << ": aiou.numRead = " << aiou->numRead <<
+		std::cerr << "INFO: R_" << whoami << ": aiou.numRead = " << aiou->numRead <<
 			" aiou.numWrite = " << aiou->numWrite << std::endl;
 
 	// release handles (broadcast channel)
 	bP_in.clear(), bP_out.clear(), bP_key.clear();
 	if (opt_verbose)
-		std::cout << "R_" << whoami << ": aiou2.numRead = " << aiou2->numRead <<
+		std::cerr << "INFO: R_" << whoami << ": aiou2.numRead = " << aiou2->numRead <<
 			" aiou2.numWrite = " << aiou2->numWrite << std::endl;
 
 	// release asynchronous unicast and broadcast
@@ -502,7 +516,7 @@ void fork_instance
 	(const size_t whoami)
 {
 	if ((pid[whoami] = fork()) < 0)
-		perror("dkg-revoke (fork)");
+		perror("ERROR: dkg-revoke (fork)");
 	else
 	{
 		if (pid[whoami] == 0)
@@ -515,14 +529,14 @@ void fork_instance
 			run_instance(whoami, sigtime, opt_r, 0);
 #endif
 			if (opt_verbose)
-				std::cout << "R_" << whoami << ": exit(0)" << std::endl;
+				std::cerr << "INFO: R_" << whoami << ": exit(0)" << std::endl;
 			exit(0);
 			/* END child code: participant R_i */
 		}
 		else
 		{
 			if (opt_verbose)
-				std::cout << "fork() = " << pid[whoami] << std::endl;
+				std::cerr << "INFO: fork() = " << pid[whoami] << std::endl;
 			instance_forked = true;
 		}
 	}
@@ -717,9 +731,27 @@ int main
 	}
 	if (opt_verbose)
 	{
-		std::cout << "INFO: canonicalized peer list = " << std::endl;
+		std::cerr << "INFO: canonicalized peer list = " << std::endl;
 		for (size_t i = 0; i < peers.size(); i++)
-			std::cout << peers[i] << std::endl;
+			std::cerr << peers[i] << std::endl;
+	}
+
+	// lock memory
+	if (!lock_memory())
+	{
+		std::cerr << "WARNING: locking memory failed; CAP_IPC_LOCK required for memory protection" << std::endl;
+		// at least try to use libgcrypt's secure memory
+		if (!gcry_check_version(TMCG_LIBGCRYPT_VERSION))
+		{
+			std::cerr << "ERROR: libgcrypt version >= " << TMCG_LIBGCRYPT_VERSION << " required" << std::endl;
+			return -1;
+		}
+		gcry_control(GCRYCTL_SUSPEND_SECMEM_WARN);
+		gcry_control(GCRYCTL_USE_SECURE_RNDPOOL);
+		gcry_control(GCRYCTL_INIT_SECMEM, 16384, 0);
+		gcry_control(GCRYCTL_RESUME_SECMEM_WARN);
+		gcry_control(GCRYCTL_INITIALIZATION_FINISHED, 0);
+		libgcrypt_secmem = true;
 	}
 
 	// initialize LibTMCG
@@ -729,7 +761,7 @@ int main
 		return -1;
 	}
 	if (opt_verbose)
-		std::cout << "INFO: using LibTMCG version " << version_libTMCG() << std::endl;
+		std::cerr << "INFO: using LibTMCG version " << version_libTMCG() << std::endl;
 	
 	// initialize return code
 	int ret = 0;
@@ -815,16 +847,16 @@ int main
 	std::cerr << "WARNING: GNUnet CADET is required for the message exchange of this program" << std::endl;
 #endif
 
-	std::cout << "INFO: running local test with " << peers.size() << " participants" << std::endl;
+	std::cerr << "INFO: running local test with " << peers.size() << " participants" << std::endl;
 	// open pipes
 	for (size_t i = 0; i < peers.size(); i++)
 	{
 		for (size_t j = 0; j < peers.size(); j++)
 		{
 			if (pipe(pipefd[i][j]) < 0)
-				perror("dkg-revoke (pipe)");
+				perror("ERROR: dkg-revoke (pipe)");
 			if (pipe(broadcast_pipefd[i][j]) < 0)
-				perror("dkg-revoke (pipe)");
+				perror("ERROR: dkg-revoke (pipe)");
 		}
 	}
 	
@@ -840,9 +872,9 @@ int main
 	{
 		int wstatus = 0;
 		if (opt_verbose)
-			std::cout << "waitpid(" << pid[i] << ")" << std::endl;
+			std::cerr << "INFO: waitpid(" << pid[i] << ")" << std::endl;
 		if (waitpid(pid[i], &wstatus, 0) != pid[i])
-			perror("dkg-revoke (waitpid)");
+			perror("ERROR: dkg-revoke (waitpid)");
 		if (!WIFEXITED(wstatus))
 		{
 			std::cerr << "ERROR: protocol instance ";
@@ -855,16 +887,16 @@ int main
 		else if (WIFEXITED(wstatus))
 		{
 			if (opt_verbose)
-				std::cout << "INFO: protocol instance " << pid[i] << " terminated with exit status " << WEXITSTATUS(wstatus) << std::endl;
+				std::cerr << "INFO: protocol instance " << pid[i] << " terminated with exit status " << WEXITSTATUS(wstatus) << std::endl;
 			if (WEXITSTATUS(wstatus))
 				ret = -2; // error
 		}
 		for (size_t j = 0; j < peers.size(); j++)
 		{
 			if ((close(pipefd[i][j][0]) < 0) || (close(pipefd[i][j][1]) < 0))
-				perror("dkg-revoke (close)");
+				perror("ERROR: dkg-revoke (close)");
 			if ((close(broadcast_pipefd[i][j][0]) < 0) || (close(broadcast_pipefd[i][j][1]) < 0))
-				perror("dkg-revoke (close)");
+				perror("ERROR: dkg-revoke (close)");
 		}
 	}
 	
