@@ -1,7 +1,7 @@
 /*******************************************************************************
    This file is part of Distributed Privacy Guard (DKGPG).
 
- Copyright (C) 2017  Heiko Stamer <HeikoStamer@gmx.net>
+ Copyright (C) 2017, 2018  Heiko Stamer <HeikoStamer@gmx.net>
 
    DKGPG is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -49,14 +49,15 @@ static const char *about = PACKAGE_STRING " " PACKAGE_URL;
 #include "dkg-tcpip-common.hh"
 #include "dkg-gnunet-common.hh"
 #include "dkg-common.hh"
+#include "dkg-io.hh"
 
 int 					pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2], broadcast_pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
 pid_t 					pid[DKGPG_MAX_N];
 std::vector<std::string>		peers;
 bool					instance_forked = false;
 
-std::string				passphrase, userid, ifilename, ofilename, passwords, hostname, port;
-tmcg_octets_t				keyid, subkeyid, pub, sub, uidsig, subsig, sec, ssb, uid;
+std::string				passphrase, userid, ifilename, ofilename, passwords, hostname, port, URI;
+tmcg_openpgp_octets_t			keyid, subkeyid, pub, sub, uidsig, subsig, sec, ssb, uid;
 std::map<size_t, size_t>		idx2dkg, dkg2idx;
 mpz_t					dss_p, dss_q, dss_g, dss_h, dss_x_i, dss_xprime_i, dss_y;
 size_t					dss_n, dss_t, dss_i;
@@ -68,13 +69,17 @@ std::vector<size_t>			dkg_qual;
 std::vector<mpz_ptr>			dkg_v_i;
 std::vector< std::vector<mpz_ptr> >	dkg_c_ik;
 gcry_mpi_t 				dsa_p, dsa_q, dsa_g, dsa_y, dsa_x, elg_p, elg_q, elg_g, elg_y, elg_x;
+gcry_mpi_t				dsa_r, dsa_s, elg_r, elg_s, rsa_n, rsa_e, rsa_md;
 gcry_mpi_t 				gk, myk, sig_r, sig_s;
+gcry_mpi_t				revdsa_r, revdsa_s, revelg_r, revelg_s, revrsa_md;
 
 int 					opt_verbose = 0;
+bool					libgcrypt_secmem = false;
 char					*opt_ifilename = NULL;
 char					*opt_ofilename = NULL;
 char					*opt_passwords = NULL;
 char					*opt_hostname = NULL;
+char					*opt_URI = NULL;
 unsigned long int			opt_e = 0, opt_p = 55000, opt_W = 5;
 
 void run_instance
@@ -82,6 +87,12 @@ void run_instance
 {
 	// read and parse the private key
 	std::string armored_seckey, thispeer = peers[whoami];
+	if (!check_strict_permissions(thispeer + "_dkg-sec.asc"))
+	{
+		std::cerr << "WARNING: weak permissions of private key file detected" << std::endl;
+		if (!set_strict_permissions(thispeer + "_dkg-sec.asc"))
+			exit(-1);
+	}
 	if (!read_key_file(thispeer + "_dkg-sec.asc", armored_seckey))
 		exit(-1);
 	init_mpis();
@@ -97,13 +108,15 @@ void run_instance
 #ifdef DKGPG_TESTSUITE
 		passphrase = "Test";
 #else
-		std::cout << "Please enter the passphrase to unlock your private key: ";
-		std::getline(std::cin, passphrase);
-		std::cin.clear();
+		if (!get_passphrase("Please enter passphrase to unlock your private key", passphrase))
+		{
+			release_mpis();
+			exit(-1);
+		}
 #endif
 		if (!parse_private_key(armored_seckey, ckeytime, ekeytime, CAPL))
 		{
-			std::cerr << "S_" << whoami << ": wrong passphrase to unlock private key" << std::endl;
+			std::cerr << "ERROR: S_" << whoami << ": wrong passphrase to unlock private key" << std::endl;
 			release_mpis();
 			exit(-1);
 		}
@@ -120,14 +133,14 @@ void run_instance
 			std::string pwd;
 			if (!TMCG_ParseHelper::gs(passwords, '/', pwd))
 			{
-				std::cerr << "S_" << whoami << ": " << "cannot read password for protecting channel to S_" << i << std::endl;
+				std::cerr << "ERROR: S_" << whoami << ": " << "cannot read password for protecting channel to S_" << i << std::endl;
 				release_mpis();
 				exit(-1);
 			}
 			key << pwd;
 			if (((i + 1) < peers.size()) && !TMCG_ParseHelper::nx(passwords, '/'))
 			{
-				std::cerr << "S_" << whoami << ": " << "cannot skip to next password for protecting channel to S_" << (i + 1) << std::endl;
+				std::cerr << "ERROR: S_" << whoami << ": " << "cannot skip to next password for protecting channel to S_" << (i + 1) << std::endl;
 				release_mpis();
 				exit(-1);
 			}
@@ -161,22 +174,22 @@ void run_instance
 	{
 		mpz_t xtest;
 		mpz_init_set_ui(xtest, i);
-		std::cout << "S_" << whoami << ": xtest = " << xtest << " <-> ";
+		std::cerr << "INFO: S_" << whoami << ": xtest = " << xtest << " <-> ";
 		rbc->Broadcast(xtest);
 		for (size_t ii = 0; ii < peers.size(); ii++)
 		{
 			if (!rbc->DeliverFrom(xtest, ii))
-				std::cout << "<X> ";
+				std::cerr << "<X> ";
 			else
-				std::cout << xtest << " ";
+				std::cerr << xtest << " ";
 		}
-		std::cout << std::endl;
+		std::cerr << std::endl;
 		mpz_clear(xtest);
 	}
 
 	// participants must agree on a common signature creation time (OpenPGP)
 	if (opt_verbose)
-		std::cout << "agree on a signature creation time for OpenPGP" << std::endl;
+		std::cerr << "INFO: agree on a signature creation time for OpenPGP" << std::endl;
 	time_t csigtime = 0;
 	std::vector<time_t> tvs;
 	mpz_t mtv;
@@ -195,7 +208,7 @@ void run_instance
 			}
 			else
 			{
-				std::cerr << "S_" << whoami << ": WARNING - no signature creation time stamp received from " << i << std::endl;
+				std::cerr << "WARNING: S_" << whoami << ": no signature creation time stamp received from S_" << i << std::endl;
 			}
 		}
 	}
@@ -203,26 +216,26 @@ void run_instance
 	std::sort(tvs.begin(), tvs.end());
 	if (tvs.size() < (peers.size() - T_RBC))
 	{
-		std::cerr << "S_" << whoami << ": not enough timestamps received" << std::endl;
+		std::cerr << "ERROR: S_" << whoami << ": not enough timestamps received" << std::endl;
 		delete rbc, delete aiou, delete aiou2;
 		release_mpis();
 		exit(-1);
 	}
 	csigtime = tvs[tvs.size()/2]; // use a median value as some kind of gentle agreement
 	if (opt_verbose)
-		std::cout << "S_" << whoami << ": canonicalized signature creation time = " << csigtime << std::endl;
+		std::cerr << "INFO: S_" << whoami << ": canonicalized signature creation time = " << csigtime << std::endl;
 
 	// select hash algorithm for OpenPGP based on |q| (size in bit)
-	tmcg_byte_t hashalgo = 0;
+	tmcg_openpgp_hashalgo_t hashalgo = TMCG_OPENPGP_HASHALGO_UNKNOWN;
 	if (mpz_sizeinbase(dss_q, 2L) == 256)
-		hashalgo = 8; // SHA256 (alg 8)
+		hashalgo = TMCG_OPENPGP_HASHALGO_SHA256; // SHA256 (alg 8)
 	else if (mpz_sizeinbase(dss_q, 2L) == 384)
-		hashalgo = 9; // SHA384 (alg 9)
+		hashalgo = TMCG_OPENPGP_HASHALGO_SHA384; // SHA384 (alg 9)
 	else if (mpz_sizeinbase(dss_q, 2L) == 512)
-		hashalgo = 10; // SHA512 (alg 10)
+		hashalgo = TMCG_OPENPGP_HASHALGO_SHA512; // SHA512 (alg 10)
 	else
 	{
-		std::cerr << "S_" << whoami << ": selecting hash algorithm failed for |q| = " << mpz_sizeinbase(dss_q, 2L) << std::endl;
+		std::cerr << "ERROR: S_" << whoami << ": selecting hash algorithm failed for |q| = " << mpz_sizeinbase(dss_q, 2L) << std::endl;
 		delete rbc, delete aiou, delete aiou2;
 		release_mpis();
 		exit(-1);
@@ -230,12 +243,12 @@ void run_instance
 
 	// compute the hash of the input file
 	if (opt_verbose)
-		std::cout << "hashing the input file \"" << opt_ifilename << "\"" << std::endl;
-	tmcg_octets_t trailer, hash, left;
-	CallasDonnerhackeFinneyShawThayerRFC4880::PacketSigPrepareDetachedSignature(0x00, hashalgo, csigtime, sigexptime, keyid, trailer);
+		std::cerr << "INFO: hashing the input file \"" << opt_ifilename << "\"" << std::endl;
+	tmcg_openpgp_octets_t trailer, hash, left;
+	CallasDonnerhackeFinneyShawThayerRFC4880::PacketSigPrepareDetachedSignature(0x00, hashalgo, csigtime, sigexptime, URI, keyid, trailer);
 	if (!CallasDonnerhackeFinneyShawThayerRFC4880::BinaryDocumentHash(opt_ifilename, trailer, hashalgo, hash, left))
 	{
-		std::cerr << "S_" << whoami << ": BinaryDocumentHash() failed; cannot process input file \"" << opt_ifilename << "\"" << std::endl;
+		std::cerr << "ERROR: S_" << whoami << ": BinaryDocumentHash() failed; cannot process input file \"" << opt_ifilename << "\"" << std::endl;
 		delete rbc, delete aiou, delete aiou2;
 		release_mpis();
 		exit(-1);
@@ -283,35 +296,33 @@ void run_instance
 			dss_in << dss_c_ik[i][k] << std::endl;
 	}
 	if (opt_verbose)
-		std::cout << "CanettiGennaroJareckiKrawczykRabinDSS(in, ...)" << std::endl;
+		std::cerr << "INFO: CanettiGennaroJareckiKrawczykRabinDSS(in, ...)" << std::endl;
 	CanettiGennaroJareckiKrawczykRabinDSS *dss = new CanettiGennaroJareckiKrawczykRabinDSS(dss_in);
 	if (!dss->CheckGroup())
 	{
-		std::cerr << "S_" << whoami << ": " << "tDSS domain parameters are not correctly generated!" << std::endl;
+		std::cerr << "ERROR: S_" << whoami << ": " << "tDSS domain parameters are not correctly generated!" << std::endl;
 		delete dss, delete rbc, delete aiou, delete aiou2;
 		release_mpis();
 		exit(-1);
 	}
 
 	// sign the hash
-	tmcg_byte_t buffer[1024];
-	gcry_mpi_t h, r, s;
+	tmcg_openpgp_byte_t buffer[1024];
+	gcry_mpi_t r, s, h;
 	mpz_t dsa_m, dsa_r, dsa_s;
 	size_t buflen = 0;
 	gcry_error_t ret;
+	memset(buffer, 0, sizeof(buffer));
 	for (size_t i = 0; ((i < hash.size()) && (i < sizeof(buffer))); i++, buflen++)
 		buffer[i] = hash[i];
-	h = gcry_mpi_new(2048);
 	r = gcry_mpi_new(2048);
 	s = gcry_mpi_new(2048);
 	mpz_init(dsa_m), mpz_init(dsa_r), mpz_init(dsa_s);
 	ret = gcry_mpi_scan(&h, GCRYMPI_FMT_USG, buffer, buflen, NULL);
 	if (ret)
 	{
-		std::cerr << "S_" << whoami << ": gcry_mpi_scan() failed for h" << std::endl;
-		gcry_mpi_release(h);
-		gcry_mpi_release(r);
-		gcry_mpi_release(s);
+		std::cerr << "ERROR: S_" << whoami << ": gcry_mpi_scan() failed for h" << std::endl;
+		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		delete dss, delete rbc, delete aiou, delete aiou2;
 		release_mpis();
@@ -319,65 +330,56 @@ void run_instance
 	}
 	if (!mpz_set_gcry_mpi(h, dsa_m))
 	{
-		std::cerr << "S_" << whoami << ": mpz_set_gcry_mpi() failed for dsa_m" << std::endl;
-		gcry_mpi_release(h);
-		gcry_mpi_release(r);
-		gcry_mpi_release(s);
+		std::cerr << "ERROR: S_" << whoami << ": mpz_set_gcry_mpi() failed for dsa_m" << std::endl;
+		gcry_mpi_release(r), gcry_mpi_release(s), gcry_mpi_release(h);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		delete dss, delete rbc, delete aiou, delete aiou2;
 		release_mpis();
 		exit(-1);
 	}
+	gcry_mpi_release(h);
 	std::stringstream err_log_sign;
 	if (opt_verbose)
-		std::cout << "S_" << whoami << ": dss.Sign()" << std::endl;
+		std::cerr << "INFO: S_" << whoami << ": dss.Sign()" << std::endl;
 	if (!dss->Sign(peers.size(), whoami, dsa_m, dsa_r, dsa_s, idx2dkg, dkg2idx, aiou, rbc, err_log_sign))
 	{
-		std::cerr << "S_" << whoami << ": " << "tDSS Sign() failed" << std::endl;
-		std::cerr << "S_" << whoami << ": log follows " << std::endl << err_log_sign.str();
-		gcry_mpi_release(h);
-		gcry_mpi_release(r);
-		gcry_mpi_release(s);
+		std::cerr << "ERROR: S_" << whoami << ": " << "tDSS Sign() failed" << std::endl;
+		std::cerr << "ERROR: S_" << whoami << ": log follows " << std::endl << err_log_sign.str();
+		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		delete dss, delete rbc, delete aiou, delete aiou2;
 		release_mpis();
 		exit(-1);
 	}
 	if (opt_verbose > 1)
-		std::cout << "S_" << whoami << ": log follows " << std::endl << err_log_sign.str();
-	if (!mpz_get_gcry_mpi(&r, dsa_r))
+		std::cerr << "INFO: S_" << whoami << ": log follows " << std::endl << err_log_sign.str(); 
+	if (!mpz_get_gcry_mpi(r, dsa_r))
 	{
-		std::cerr << "S_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_r" << std::endl;
-		gcry_mpi_release(h);
-		gcry_mpi_release(r);
-		gcry_mpi_release(s);
+		std::cerr << "ERROR: S_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_r" << std::endl;
+		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		delete dss, delete rbc, delete aiou, delete aiou2;
 		release_mpis();
 		exit(-1);
 	}
-	if (!mpz_get_gcry_mpi(&s, dsa_s))
+	if (!mpz_get_gcry_mpi(s, dsa_s))
 	{
-		std::cerr << "S_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_s" << std::endl;
-		gcry_mpi_release(h);
-		gcry_mpi_release(r);
-		gcry_mpi_release(s);
+		std::cerr << "ERROR: S_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_s" << std::endl;
+		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		delete dss, delete rbc, delete aiou, delete aiou2;
 		release_mpis();
 		exit(-1);
 	}
-	tmcg_octets_t sig;
+	tmcg_openpgp_octets_t sig;
 	CallasDonnerhackeFinneyShawThayerRFC4880::PacketSigEncode(trailer, left, r, s, sig);
-	gcry_mpi_release(h);
-	gcry_mpi_release(r);
-	gcry_mpi_release(s);
+	gcry_mpi_release(r), gcry_mpi_release(s);
 	mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 
 	// at the end: deliver some more rounds for still waiting parties
 	time_t synctime = aiounicast::aio_timeout_very_long;
 	if (opt_verbose)
-		std::cout << "S_" << whoami << ": waiting approximately " << (synctime * (T_RBC + 1)) << " seconds for stalled parties" << std::endl;
+		std::cerr << "INFO: S_" << whoami << ": waiting approximately " << (synctime * (T_RBC + 1)) << " seconds for stalled parties" << std::endl;
 	rbc->Sync(synctime);
 
 	// release tDSS
@@ -389,13 +391,13 @@ void run_instance
 	// release handles (unicast channel)
 	uP_in.clear(), uP_out.clear(), uP_key.clear();
 	if (opt_verbose)
-		std::cout << "S_" << whoami << ": aiou.numRead = " << aiou->numRead <<
+		std::cerr << "INFO: S_" << whoami << ": aiou.numRead = " << aiou->numRead <<
 			" aiou.numWrite = " << aiou->numWrite << std::endl;
 
 	// release handles (broadcast channel)
 	bP_in.clear(), bP_out.clear(), bP_key.clear();
 	if (opt_verbose)
-		std::cout << "S_" << whoami << ": aiou2.numRead = " << aiou2->numRead <<
+		std::cerr << "INFO: S_" << whoami << ": aiou2.numRead = " << aiou2->numRead <<
 			" aiou2.numWrite = " << aiou2->numWrite << std::endl;
 
 	// release asynchronous unicast and broadcast
@@ -406,7 +408,7 @@ void run_instance
 
 	// output the result
 	std::string sigstr;
-	CallasDonnerhackeFinneyShawThayerRFC4880::ArmorEncode(2, sig, sigstr);
+	CallasDonnerhackeFinneyShawThayerRFC4880::ArmorEncode(TMCG_OPENPGP_ARMOR_SIGNATURE, sig, sigstr);
 	if (opt_ofilename != NULL)
 	{
 		if (!write_message(opt_ofilename, sigstr))
@@ -422,6 +424,7 @@ char *gnunet_opt_ifilename = NULL;
 char *gnunet_opt_ofilename = NULL;
 char *gnunet_opt_passwords = NULL;
 char *gnunet_opt_port = NULL;
+char *gnunet_opt_URI = NULL;
 unsigned int gnunet_opt_sigexptime = 0;
 unsigned int gnunet_opt_xtests = 0;
 unsigned int gnunet_opt_wait = 5;
@@ -433,7 +436,7 @@ void fork_instance
 	(const size_t whoami)
 {
 	if ((pid[whoami] = fork()) < 0)
-		perror("dkg-sign (fork)");
+		perror("ERROR: dkg-sign (fork)");
 	else
 	{
 		if (pid[whoami] == 0)
@@ -446,14 +449,14 @@ void fork_instance
 			run_instance(whoami, sigtime, opt_e, 0);
 #endif
 			if (opt_verbose)
-				std::cout << "S_" << whoami << ": exit(0)" << std::endl;
+				std::cerr << "INFO: S_" << whoami << ": exit(0)" << std::endl;
 			exit(0);
 			/* END child code: participant S_i */
 		}
 		else
 		{
 			if (opt_verbose)
-				std::cout << "fork() = " << pid[whoami] << std::endl;
+				std::cerr << "INFO: fork() = " << pid[whoami] << std::endl;
 			instance_forked = true;
 		}
 	}
@@ -508,6 +511,12 @@ int main
 			"exchanged passwords to protect private and broadcast channels",
 			&gnunet_opt_passwords
 		),
+		GNUNET_GETOPT_option_string('U',
+			"URI",
+			"STRING",
+			"policy URI tied to signature",
+			&gnunet_opt_URI
+		),
 		GNUNET_GETOPT_option_version(version),
 		GNUNET_GETOPT_option_flag('V',
 			"verbose",
@@ -552,12 +561,16 @@ int main
 		opt_hostname = gnunet_opt_hostname;
 	if (gnunet_opt_passwords != NULL)
 		opt_passwords = gnunet_opt_passwords;
+	if (gnunet_opt_URI != NULL)
+		opt_URI = gnunet_opt_URI;
 	if (gnunet_opt_passwords != NULL)
 		passwords = gnunet_opt_passwords; // get passwords from GNUnet options
 	if (gnunet_opt_hostname != NULL)
 		hostname = gnunet_opt_hostname; // get hostname from GNUnet options
 	if (gnunet_opt_W != opt_W)
 		opt_W = gnunet_opt_W; // get aiou message timeout from GNUnet options
+	if (gnunet_opt_URI != NULL)
+		URI = gnunet_opt_URI; // get policy URI from GNUnet options
 #endif
 
 	// create peer list from remaining arguments
@@ -566,8 +579,9 @@ int main
 		std::string arg = argv[i+1];
 		// ignore options
 		if ((arg.find("-c") == 0) || (arg.find("-p") == 0) || (arg.find("-w") == 0) || (arg.find("-W") == 0) || 
-			(arg.find("-L") == 0) || (arg.find("-l") == 0) || (arg.find("-i") == 0) || (arg.find("-o") == 0) || 
-			(arg.find("-e") == 0) || (arg.find("-x") == 0) || (arg.find("-P") == 0) || (arg.find("-H") == 0))
+		    (arg.find("-L") == 0) || (arg.find("-l") == 0) || (arg.find("-i") == 0) || (arg.find("-o") == 0) || 
+		    (arg.find("-e") == 0) || (arg.find("-x") == 0) || (arg.find("-P") == 0) || (arg.find("-H") == 0) ||
+		    (arg.find("-U") == 0))
 		{
 			size_t idx = ++i;
 			if ((arg.find("-i") == 0) && (idx < (size_t)(argc - 1)) && (opt_ifilename == NULL))
@@ -590,6 +604,11 @@ int main
 				passwords = argv[i+1];
 				opt_passwords = (char*)passwords.c_str();
 			}
+			if ((arg.find("-U") == 0) && (idx < (size_t)(argc - 1)) && (opt_URI == NULL))
+			{
+				URI = argv[i+1];
+				opt_URI = (char*)URI.c_str();
+			}
 			if ((arg.find("-e") == 0) && (idx < (size_t)(argc - 1)) && (opt_e == 0))
 				opt_e = strtoul(argv[i+1], NULL, 10);
 			if ((arg.find("-p") == 0) && (idx < (size_t)(argc - 1)) && (port.length() == 0))
@@ -613,6 +632,7 @@ int main
 				std::cout << "  -o FILENAME    write detached signature to FILENAME" << std::endl;
 				std::cout << "  -p INTEGER     start port for built-in TCP/IP message exchange service" << std::endl;
 				std::cout << "  -P STRING      exchanged passwords to protect private and broadcast channels" << std::endl;
+				std::cout << "  -U STRING      policy URI tied to generated signatures" << std::endl;
 				std::cout << "  -v, --version  print the version number" << std::endl;
 				std::cout << "  -V, --verbose  turn on verbose output" << std::endl;
 				std::cout << "  -W TIME        timeout for point-to-point messages in minutes" << std::endl;
@@ -650,31 +670,15 @@ int main
 	peers.push_back("Test2");
 	peers.push_back("Test3");
 	peers.push_back("Test4");
-	ifilename = "Test1_output.asc";
+	ifilename = "Test1_output.bin";
 	opt_ifilename = (char*)ifilename.c_str();
 	ofilename = "Test1_output.sig";
 	opt_ofilename = (char*)ofilename.c_str();
+	URI = "https://savannah.nongnu.org/projects/dkgpg/";
 	opt_verbose = 1;
 #endif
-	if (peers.size() < 1)
-	{
-		std::cerr << "ERROR: no peers given as argument; usage: " << usage << std::endl;
-		return -1;
-	}
-	// canonicalize peer list
-	std::sort(peers.begin(), peers.end());
-	std::vector<std::string>::iterator it = std::unique(peers.begin(), peers.end());
-	peers.resize(std::distance(peers.begin(), it));
-	if ((peers.size() < 3)  || (peers.size() > DKGPG_MAX_N))
-	{
-		std::cerr << "ERROR: too few or too many peers given" << std::endl;
-		return -1;
-	}
-	if (!init_libTMCG())
-	{
-		std::cerr << "ERROR: initialization of LibTMCG failed" << std::endl;
-		return -1;
-	}
+
+	// check command line arguments
 	if ((opt_hostname != NULL) && (opt_passwords == NULL))
 	{
 		std::cerr << "ERROR: option \"-P\" is necessary due to insecure network" << std::endl;
@@ -685,12 +689,56 @@ int main
 		std::cerr << "ERROR: option -i required to specify an input file" << std::endl;
 		return -1;
 	}
+	if (peers.size() < 1)
+	{
+		std::cerr << "ERROR: no peers given as argument; usage: " << usage << std::endl;
+		return -1;
+	}
+
+	// canonicalize peer list
+	std::sort(peers.begin(), peers.end());
+	std::vector<std::string>::iterator it = std::unique(peers.begin(), peers.end());
+	peers.resize(std::distance(peers.begin(), it));
+	if ((peers.size() < 3)  || (peers.size() > DKGPG_MAX_N))
+	{
+		std::cerr << "ERROR: too few or too many peers given" << std::endl;
+		return -1;
+	}
 	if (opt_verbose)
 	{
-		std::cout << "INFO: canonicalized peer list = " << std::endl;
+		std::cerr << "INFO: canonicalized peer list = " << std::endl;
 		for (size_t i = 0; i < peers.size(); i++)
-			std::cout << peers[i] << std::endl;
+			std::cerr << peers[i] << std::endl;
 	}
+
+	// lock memory
+	if (!lock_memory())
+	{
+		std::cerr << "WARNING: locking memory failed; CAP_IPC_LOCK required for memory protection" << std::endl;
+		// at least try to use libgcrypt's secure memory
+		if (!gcry_check_version(TMCG_LIBGCRYPT_VERSION))
+		{
+			std::cerr << "ERROR: libgcrypt version >= " << TMCG_LIBGCRYPT_VERSION << " required" << std::endl;
+			return -1;
+		}
+		gcry_control(GCRYCTL_SUSPEND_SECMEM_WARN);
+		gcry_control(GCRYCTL_USE_SECURE_RNDPOOL);
+		gcry_control(GCRYCTL_INIT_SECMEM, 16384, 0);
+		gcry_control(GCRYCTL_RESUME_SECMEM_WARN);
+		gcry_control(GCRYCTL_INITIALIZATION_FINISHED, 0);
+		libgcrypt_secmem = true;
+	}
+
+	// initialize LibTMCG
+	if (!init_libTMCG())
+	{
+		std::cerr << "ERROR: initialization of LibTMCG failed" << std::endl;
+		return -1;
+	}
+	if (opt_verbose)
+		std::cerr << "INFO: using LibTMCG version " << version_libTMCG() << std::endl;
+	
+	
 	// initialize return code
 	int ret = 0;
 	// create underlying point-to-point channels, if built-in TCP/IP service requested
@@ -752,6 +800,12 @@ int main
 			"exchanged passwords to protect private and broadcast channels",
 			&gnunet_opt_passwords
 		),
+		GNUNET_GETOPT_option_string('U',
+			"URI",
+			"STRING",
+			"policy URI tied to signature",
+			&gnunet_opt_URI
+		),
 		GNUNET_GETOPT_option_flag('V',
 			"verbose",
 			"turn on verbose output",
@@ -787,16 +841,16 @@ int main
 	std::cerr << "WARNING: GNUnet CADET is required for the message exchange of this program" << std::endl;
 #endif
 
-	std::cout << "INFO: running local test with " << peers.size() << " participants" << std::endl;
+	std::cerr << "INFO: running local test with " << peers.size() << " participants" << std::endl;
 	// open pipes
 	for (size_t i = 0; i < peers.size(); i++)
 	{
 		for (size_t j = 0; j < peers.size(); j++)
 		{
 			if (pipe(pipefd[i][j]) < 0)
-				perror("dkg-sign (pipe)");
+				perror("ERROR: dkg-sign (pipe)");
 			if (pipe(broadcast_pipefd[i][j]) < 0)
-				perror("dkg-sign (pipe)");
+				perror("ERROR: dkg-sign (pipe)");
 		}
 	}
 	
@@ -812,9 +866,9 @@ int main
 	{
 		int wstatus = 0;
 		if (opt_verbose)
-			std::cout << "waitpid(" << pid[i] << ")" << std::endl;
+			std::cerr << "INFO: waitpid(" << pid[i] << ")" << std::endl;
 		if (waitpid(pid[i], &wstatus, 0) != pid[i])
-			perror("dkg-sign (waitpid)");
+			perror("ERROR: dkg-sign (waitpid)");
 		if (!WIFEXITED(wstatus))
 		{
 			std::cerr << "ERROR: protocol instance ";
@@ -827,16 +881,16 @@ int main
 		else if (WIFEXITED(wstatus))
 		{
 			if (opt_verbose)
-				std::cout << "INFO: protocol instance " << pid[i] << " terminated with exit status " << WEXITSTATUS(wstatus) << std::endl;
+				std::cerr << "INFO: protocol instance " << pid[i] << " terminated with exit status " << WEXITSTATUS(wstatus) << std::endl;
 			if (WEXITSTATUS(wstatus))
 				ret = -2; // error
 		}
 		for (size_t j = 0; j < peers.size(); j++)
 		{
 			if ((close(pipefd[i][j][0]) < 0) || (close(pipefd[i][j][1]) < 0))
-				perror("dkg-sign (close)");
+				perror("ERROR: dkg-sign (close)");
 			if ((close(broadcast_pipefd[i][j][0]) < 0) || (close(broadcast_pipefd[i][j][1]) < 0))
-				perror("dkg-sign (close)");
+				perror("ERROR: dkg-sign (close)");
 		}
 	}
 	
