@@ -1,7 +1,7 @@
 /*******************************************************************************
    This file is part of Distributed Privacy Guard (DKGPG).
 
- Copyright (C) 2017  Heiko Stamer <HeikoStamer@gmx.net>
+ Copyright (C) 2017, 2018  Heiko Stamer <HeikoStamer@gmx.net>
 
    DKGPG is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -35,7 +35,7 @@ int main
 	static const char *about = PACKAGE_STRING " " PACKAGE_URL;
 	static const char *version = PACKAGE_VERSION " (" PACKAGE_NAME ")";
 	size_t factor = 0;
-	std::string fips;
+	std::string fips, prefix;
 	int opt_verbose = 0;
 
 	for (size_t i = 0; i < (size_t)(argc - 1); i++)
@@ -49,6 +49,13 @@ int main
 				fips = argv[i+1];
 			continue;
 		}
+		else if ((arg.find("-k") == 0))
+		{
+			size_t idx = ++i;
+			if ((arg.find("-k") == 0) && (idx < (size_t)(argc - 1)) && (!prefix.length()))
+				prefix = argv[i+1];
+			continue;
+		}
 		else if ((arg.find("--") == 0) || (arg.find("-v") == 0) || (arg.find("-h") == 0) || (arg.find("-V") == 0))
 		{
 			if ((arg.find("-h") == 0) || (arg.find("--help") == 0))
@@ -58,6 +65,7 @@ int main
 				std::cout << "Arguments mandatory for long options are also mandatory for short options." << std::endl;
 				std::cout << "  -h, --help     print this help" << std::endl;
 				std::cout << "  -f SEED        generate domain parameters according to FIPS 186-4 with SEED" << std::endl;
+				std::cout << "  -k PREFIX      generate value k with given PREFIX (not in FIPS mode)" << std::endl;
 				std::cout << "  -v, --version  print the version number" << std::endl;
 				std::cout << "  -V, --verbose  turn on verbose output" << std::endl;
 				return 0; // not continue
@@ -81,7 +89,8 @@ int main
 
 #ifdef DKGPG_TESTSUITE
 	factor = 1;
-	fips = "DKGPGTESTSUITEDKGPGTESTSUITEDKGPGTESTSUITEDKGPGTESTSUITEDKGPGTESTSUITE";
+	if (mpz_wrandom_ui() % 2)
+		fips = "DKGPGTESTSUITEDKGPGTESTSUITEDKGPGTESTSUITEDKGPGTESTSUITEDKGPGTESTSUITE";
 	opt_verbose = 1;
 #endif
 
@@ -92,7 +101,7 @@ int main
 		return -1;
 	}
 	if (opt_verbose)
-		std::cout << "INFO: using LibTMCG version " << version_libTMCG() << std::endl;
+		std::cerr << "INFO: using LibTMCG version " << version_libTMCG() << std::endl;
 
 	if (fips.length()) // generate primes and generator according to FIPS 186-4?
 	{
@@ -102,13 +111,13 @@ int main
 		if (factor > 0)
 		{
 			if (opt_verbose)
-				std::cout << "Generating primes p and q according to FIPS 186-4 with factor = " << factor << std::endl;
+				std::cerr << "INFO: Generating primes p and q according to FIPS 186-4 with factor = " << factor << std::endl;
 			L = TMCG_DDH_SIZE + (factor * 1024), N = TMCG_DLSE_SIZE + ((factor - 1) * 128);
 		}
 		else
 		{
 			if (opt_verbose)
-				std::cout << "Generating primes p and q according to FIPS 186-4 with default sizes" << std::endl;
+				std::cerr << "INFO: Generating primes p and q according to FIPS 186-4 with default sizes" << std::endl;
 			L = TMCG_DDH_SIZE, N = TMCG_DLSE_SIZE;
 		}
 		int hash_algo = 0, mr_iterations = 0;
@@ -163,7 +172,7 @@ int main
 				mpz_fhash(U, hash_algo, domain_parameter_seed);
 				mpz_tdiv_r_2exp(U, U, N - 1);
 				if (opt_verbose)
-					std::cout << "U = " << U << std::endl;
+					std::cerr << "INFO: U = " << U << std::endl;
 				// 7. $q = 2^{N-1} + U + 1 - (U \bmod 2)$.
 				mpz_set_ui(q, 1L);
 				mpz_mul_2exp(q, q, N - 1);
@@ -172,7 +181,7 @@ int main
 				if (mpz_odd_p(U))
 					mpz_sub_ui(q, q, 1L);
 				if (opt_verbose)
-					std::cout << "q = " << q << std::endl;
+					std::cerr << "INFO: q = " << q << std::endl;
 				// 8. Test whether or not $q$ is prime as specified in Appendix C.3.
 				// 9. If $q$ is not a prime, then go to step 5.
 				if (!mpz_probab_prime_p(q, mr_iterations))
@@ -196,8 +205,8 @@ int main
 					mpz_add_ui(tmp, tmp, j);
 					mpz_tdiv_r_2exp(tmp, tmp, seedlen);
 					mpz_fhash(V_j[j], hash_algo, tmp);
-					if (opt_verbose)
-						std::cout << "V_j[" << j << "] = " << V_j[j] << std::endl;
+					if (opt_verbose > 1)
+						std::cerr << "INFO: V_j[" << j << "] = " << V_j[j] << std::endl;
 					mpz_clear(tmp);
 				}
 				// 11.2 $W = V_0 + (V_1 * 2^{outlen}) + \cdots + (V_{n-1} * 2^{(n-1)*outlen}) + ((V_n \bmod 2^b) * 2^{n*outlen})$.
@@ -213,7 +222,7 @@ int main
 					mpz_clear(tmp);
 				}
 				if (opt_verbose)
-					std::cout << "W = " << W << std::endl;
+					std::cerr << "INFO: W = " << W << std::endl;
 				// 11.3 $X = W + 2^{L-1}$.
 				mpz_set_ui(X, 1L);
 				mpz_mul_2exp(X, X, L - 1);
@@ -242,11 +251,11 @@ int main
 				break;
 		}
 		if (opt_verbose)
-			std::cout << "p = " << p << std::endl;
+			std::cerr << "INFO: p = " << p << std::endl;
 		if (opt_verbose)
-			std::cout << "counter = " << counter << std::endl;
+			std::cerr << "INFO: counter = " << counter << std::endl;
 		if (opt_verbose)
-			std::cout << "Computing generator g according to FIPS 186-4" << std::endl;
+			std::cerr << "INFO: Computing generator g according to FIPS 186-4" << std::endl;
 		// 1. If ($index$ is incorrect), then return INVALID.
 		mpz_t index;
 		mpz_init_set_ui(index, 108L); // fixed index value for DKG-tools
@@ -293,7 +302,7 @@ int main
 			break;
 		}
 		if (opt_verbose)
-			std::cout << "g = " << p << std::endl;		
+			std::cerr << "INFO: g = " << p << std::endl;		
 
 		// export group parameters to stdout
 		mpz_t hash_algo_mpz, counter_mpz;
@@ -329,21 +338,61 @@ int main
 	{
 		// create VTMF instance as a CRS (common reference string)
 		BarnettSmartVTMF_dlog *vtmf = NULL;
-		if (factor > 0)
+		if (prefix.length())
 		{
 			if (opt_verbose)
-				std::cout << "Generating primes p and q with factor = " << factor <<
+				std::cerr << "INFO: Generating primes p and q with k-prefix = " << prefix << ", factor = " << factor <<
 					" and canonical generator g (by VTMF)" << std::endl;
-			// for each argument, sizes of underlying finite field and subgroup are increased by 1024 bit resp. 128 bit
-			vtmf = new BarnettSmartVTMF_dlog(TMCG_DDH_SIZE + (factor * 1024), TMCG_DLSE_SIZE + (factor * 128), true);
+			mpz_t p, q, g, k;
+			mpz_init(p), mpz_init(q), mpz_init(g), mpz_init(k);
+			if (mpz_set_str(k, prefix.c_str(), TMCG_MPZ_IO_BASE) < 0)
+			{
+				mpz_clear(p), mpz_clear(q), mpz_clear(g), mpz_clear(k);
+				std::cerr << "ERROR: cannot convert given PREFIX to MPI value (wrong base)" << std::endl;
+				return -1;
+			}
+			mpz_lprime_prefix(p, q, k, TMCG_DDH_SIZE + (factor * 1024), TMCG_DLSE_SIZE + (factor * 128), TMCG_MR_ITERATIONS);
+			mpz_t foo, bar;
+			mpz_init(foo), mpz_init(bar);
+			mpz_sub_ui(foo, p, 1L); // compute $p-1$
+			// We use a procedure similar to FIPS 186-3 A.2.3;
+			// it is supposed as verifiable generation of $g$.
+			std::stringstream U;
+			U << "LibTMCG|" << p << "|" << q << "|ggen|";
+			do
+			{
+				mpz_shash(bar, U.str());
+				mpz_powm(g, bar, k, p); // $g := [bar]^k \bmod p$
+				U << g << "|";
+				mpz_powm(bar, g, q, p);
+				// check $1 < g < p-1$ and $g^q \equiv 1 \pmod{p}$
+			}
+			while (!mpz_cmp_ui(g, 0L) || !mpz_cmp_ui(g, 1L) || 
+				!mpz_cmp(g, foo) || mpz_cmp_ui(bar, 1L));
+			mpz_clear(foo), mpz_clear(bar);
+			std::stringstream input;
+			input << p << std::endl << q << std::endl << g << std::endl << k << std::endl;
+			mpz_clear(p), mpz_clear(q), mpz_clear(g), mpz_clear(k);
+			vtmf = new BarnettSmartVTMF_dlog(input, TMCG_DDH_SIZE, TMCG_DLSE_SIZE, true);
 		}
 		else
 		{
-			if (opt_verbose)
-				std::cout << "Generating primes p and q with default sizes" <<
-					" and canonical generator g (by VTMF)" << std::endl;
-			// use default security parameter from LibTMCG and verifiable generation of $g$
-			vtmf = new BarnettSmartVTMF_dlog(TMCG_DDH_SIZE, TMCG_DLSE_SIZE, true);
+			if (factor > 0)
+			{
+				if (opt_verbose)
+					std::cerr << "INFO: Generating primes p and q with factor = " << factor <<
+						" and canonical generator g (by VTMF)" << std::endl;
+				// for each argument, sizes of underlying finite field and subgroup are increased by 1024 bit resp. 128 bit
+				vtmf = new BarnettSmartVTMF_dlog(TMCG_DDH_SIZE + (factor * 1024), TMCG_DLSE_SIZE + (factor * 128), true);
+			}
+			else
+			{
+				if (opt_verbose)
+					std::cerr << "INFO: Generating primes p and q with default sizes" <<
+						" and canonical generator g (by VTMF)" << std::endl;
+				// use default security parameter from LibTMCG and verifiable generation of $g$
+				vtmf = new BarnettSmartVTMF_dlog(TMCG_DDH_SIZE, TMCG_DLSE_SIZE, true);
+			}
 		}
 
 		// check the instance for sanity
