@@ -35,11 +35,12 @@
 #include <libTMCG.hh>
 
 #include "dkg-common.hh"
+#include "dkg-io.hh"
 
 std::vector<std::string>		peers;
 
 std::string				passphrase, userid;
-tmcg_octets_t				keyid, subkeyid, pub, sub, uidsig, subsig, sec, ssb, uid;
+tmcg_openpgp_octets_t			keyid, subkeyid, pub, sub, uidsig, subsig, sec, ssb, uid;
 std::map<size_t, size_t>		idx2dkg, dkg2idx;
 mpz_t					dss_p, dss_q, dss_g, dss_h, dss_x_i, dss_xprime_i, dss_y;
 size_t					dss_n, dss_t, dss_i;
@@ -168,10 +169,16 @@ int main
 		return -1;
 	}
 	if (opt_verbose)
-		std::cout << "INFO: using LibTMCG version " << version_libTMCG() << std::endl;
+		std::cerr << "INFO: using LibTMCG version " << version_libTMCG() << std::endl;
 
 	// read and parse the private key
 	std::string armored_seckey, thispeer = peers[0];
+	if (!check_strict_permissions(thispeer + "_dkg-sec.asc"))
+	{
+		std::cerr << "WARNING: weak permissions of private key file detected" << std::endl;
+		if (!set_strict_permissions(thispeer + "_dkg-sec.asc"))
+			exit(-1);
+	}
 	if (!read_key_file(thispeer + "_dkg-sec.asc", armored_seckey))
 		return -1;
 	init_mpis();
@@ -238,7 +245,7 @@ int main
 			dss_in << dss_c_ik[i][k] << std::endl;
 	}
 	if (opt_verbose)
-		std::cout << "CanettiGennaroJareckiKrawczykRabinDSS(in, ...)" << std::endl;
+		std::cerr << "INFO: CanettiGennaroJareckiKrawczykRabinDSS(in, ...)" << std::endl;
 	CanettiGennaroJareckiKrawczykRabinDSS *dss = new CanettiGennaroJareckiKrawczykRabinDSS(dss_in);
 	if (!dss->CheckGroup())
 	{
@@ -281,7 +288,7 @@ int main
 				dkg_in << dkg_c_ik[i][k] << std::endl;
 		}
 		if (opt_verbose)
-			std::cout << "GennaroJareckiKrawczykRabinDKG(in, ...)" << std::endl;
+			std::cerr << "INFO: GennaroJareckiKrawczykRabinDKG(in, ...)" << std::endl;
 		dkg = new GennaroJareckiKrawczykRabinDKG(dkg_in);
 		if (!dkg->CheckGroup())
 		{
@@ -306,7 +313,7 @@ int main
 	for (size_t i = 0; i < keyid.size(); i++)
 		std::cout << std::setfill('0') << std::setw(2) << std::right << (int)keyid[i] << " ";
 	std::cout << std::dec << std::endl;
-	tmcg_octets_t pub_hashing, fpr;
+	tmcg_openpgp_octets_t pub_hashing, fpr;
 	for (size_t i = 6; i < pub.size(); i++)
 		pub_hashing.push_back(pub[i]);
 	CallasDonnerhackeFinneyShawThayerRFC4880::FingerprintCompute(pub_hashing, fpr);
@@ -365,7 +372,7 @@ int main
 		for (size_t i = 0; i < subkeyid.size(); i++)
 			std::cout << std::setfill('0') << std::setw(2) << std::right << (int)subkeyid[i] << " ";
 		std::cout << std::dec << std::endl;
-		tmcg_octets_t sub_hashing, sub_fpr;
+		tmcg_openpgp_octets_t sub_hashing, sub_fpr;
 		for (size_t i = 6; i < sub.size(); i++)
 			sub_hashing.push_back(sub[i]);
 		CallasDonnerhackeFinneyShawThayerRFC4880::FingerprintCompute(sub_hashing, sub_fpr);
@@ -639,7 +646,7 @@ int main
 				for (size_t k = 0; k < c_ik[j].size(); k++)
 					gcry_mpi_release(c_ik[j][k]);
 			// export updated private key in OpenPGP armor format
-			tmcg_octets_t all;
+			tmcg_openpgp_octets_t all;
 			std::string armor;
 			std::stringstream secfilename;
 			secfilename << thispeer << "_dkg-sec.asc";
@@ -676,7 +683,7 @@ int main
 			}
 			secofs.close();
 			if (opt_verbose)
-				std::cout << "INFO: migration from peer \"" << migrate_peer_from << "\" to \"" << migrate_peer_to << "\" finished" << std::endl;
+				std::cerr << "INFO: migration from peer \"" << migrate_peer_from << "\" to \"" << migrate_peer_to << "\" finished" << std::endl;
 		}
 		else
 			std::cerr << "WARNING: migration not possible due to missing or bad tDSS key" << std::endl;
