@@ -1,7 +1,7 @@
 /*******************************************************************************
    This file is part of Distributed Privacy Guard (DKGPG).
 
- Copyright (C) 2017  Heiko Stamer <HeikoStamer@gmx.net>
+ Copyright (C) 2017, 2018  Heiko Stamer <HeikoStamer@gmx.net>
 
    DKGPG is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -49,6 +49,7 @@ static const char *about = PACKAGE_STRING " " PACKAGE_URL;
 #include "dkg-tcpip-common.hh"
 #include "dkg-gnunet-common.hh"
 #include "dkg-common.hh"
+#include "dkg-io.hh"
 
 int 					pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2], broadcast_pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
 pid_t 					pid[DKGPG_MAX_N];
@@ -56,7 +57,7 @@ std::vector<std::string>		peers;
 bool					instance_forked = false;
 
 std::string				passphrase, userid, passwords, hostname, port;
-tmcg_octets_t				keyid, subkeyid, pub, sub, uidsig, subsig, sec, ssb, uid;
+tmcg_openpgp_octets_t			keyid, subkeyid, pub, sub, uidsig, subsig, sec, ssb, uid;
 std::map<size_t, size_t>		idx2dkg, dkg2idx;
 mpz_t					dss_p, dss_q, dss_g, dss_h, dss_x_i, dss_xprime_i, dss_y;
 size_t					dss_n, dss_t, dss_i;
@@ -68,18 +69,30 @@ std::vector<size_t>			dkg_qual;
 std::vector<mpz_ptr>			dkg_v_i;
 std::vector< std::vector<mpz_ptr> >	dkg_c_ik;
 gcry_mpi_t 				dsa_p, dsa_q, dsa_g, dsa_y, dsa_x, elg_p, elg_q, elg_g, elg_y, elg_x;
+gcry_mpi_t				dsa_r, dsa_s, elg_r, elg_s, rsa_n, rsa_e, rsa_md;
 gcry_mpi_t 				gk, myk, sig_r, sig_s;
+gcry_mpi_t				revdsa_r, revdsa_s, revelg_r, revelg_s, revrsa_md;
 
 int 					opt_verbose = 0;
+bool					libgcrypt_secmem = false;
 char					*opt_passwords = NULL;
 char					*opt_hostname = NULL;
-unsigned long int			opt_p = 55000;
+unsigned long int			opt_p = 55000, opt_W = 5;
+
+mpz_t					cache[TMCG_MAX_SSRANDOMM_CACHE], cache_mod;
+size_t					cache_avail = 0;
 
 void run_instance
 	(size_t whoami, const size_t num_xtests)
 {
 	// read and parse the private key
 	std::string armored_seckey, thispeer = peers[whoami];
+	if (!check_strict_permissions(thispeer + "_dkg-sec.asc"))
+	{
+		std::cerr << "WARNING: weak permissions of private key file detected" << std::endl;
+		if (!set_strict_permissions(thispeer + "_dkg-sec.asc"))
+			exit(-1);
+	}
 	if (!read_key_file(thispeer + "_dkg-sec.asc", armored_seckey))
 		exit(-1);
 	init_mpis();
@@ -95,13 +108,15 @@ void run_instance
 #ifdef DKGPG_TESTSUITE
 		passphrase = "Test";
 #else
-		std::cout << "Please enter the passphrase to unlock your private key: ";
-		std::getline(std::cin, passphrase);
-		std::cin.clear();
+		if (!get_passphrase("Please enter passphrase to unlock your private key", passphrase))
+		{
+			release_mpis();
+			exit(-1);
+		}
 #endif
 		if (!parse_private_key(armored_seckey, ckeytime, ekeytime, CAPL))
 		{
-			std::cerr << "R_" << whoami << ": wrong passphrase to unlock private key" << std::endl;
+			std::cerr << "ERROR: R_" << whoami << ": wrong passphrase to unlock private key" << std::endl;
 			release_mpis();
 			exit(-1);
 		}
@@ -118,14 +133,14 @@ void run_instance
 			std::string pwd;
 			if (!TMCG_ParseHelper::gs(passwords, '/', pwd))
 			{
-				std::cerr << "R_" << whoami << ": " << "cannot read password for protecting channel to R_" << i << std::endl;
+				std::cerr << "ERROR: R_" << whoami << ": " << "cannot read password for protecting channel to R_" << i << std::endl;
 				release_mpis();
 				exit(-1);
 			}
 			key << pwd;
 			if (((i + 1) < peers.size()) && !TMCG_ParseHelper::nx(passwords, '/'))
 			{
-				std::cerr << "R_" << whoami << ": " << "cannot skip to next password for protecting channel to R_" << (i + 1) << std::endl;
+				std::cerr << "ERROR: R_" << whoami << ": " << "cannot skip to next password for protecting channel to R_" << (i + 1) << std::endl;
 				release_mpis();
 				exit(-1);
 			}
@@ -141,17 +156,17 @@ void run_instance
 	}
 
 	// create asynchronous authenticated unicast channels
-	aiounicast_select *aiou = new aiounicast_select(peers.size(), whoami, uP_in, uP_out, uP_key);
+	aiounicast_select *aiou = new aiounicast_select(peers.size(), whoami, uP_in, uP_out, uP_key, aiounicast::aio_scheduler_roundrobin, (opt_W * 60));
 
 	// create asynchronous authenticated unicast channels for broadcast protocol
-	aiounicast_select *aiou2 = new aiounicast_select(peers.size(), whoami, bP_in, bP_out, bP_key);
+	aiounicast_select *aiou2 = new aiounicast_select(peers.size(), whoami, bP_in, bP_out, bP_key, aiounicast::aio_scheduler_roundrobin, (opt_W * 60));
 			
 	// create an instance of a reliable broadcast protocol (RBC)
 	std::string myID = "dkg-refresh|";
 	for (size_t i = 0; i < peers.size(); i++)
 		myID += peers[i] + "|";
 	size_t T_RBC = (peers.size() - 1) / 3; // assume maximum asynchronous t-resilience for RBC
-	CachinKursawePetzoldShoupRBC *rbc = new CachinKursawePetzoldShoupRBC(peers.size(), T_RBC, whoami, aiou2);
+	CachinKursawePetzoldShoupRBC *rbc = new CachinKursawePetzoldShoupRBC(peers.size(), T_RBC, whoami, aiou2, aiounicast::aio_scheduler_roundrobin, (opt_W * 60));
 	rbc->setID(myID);
 
 	// perform a simple exchange test with debug output
@@ -159,16 +174,16 @@ void run_instance
 	{
 		mpz_t xtest;
 		mpz_init_set_ui(xtest, i);
-		std::cout << "R_" << whoami << ": xtest = " << xtest << " <-> ";
+		std::cerr << "INFO: R_" << whoami << ": xtest = " << xtest << " <-> ";
 		rbc->Broadcast(xtest);
 		for (size_t ii = 0; ii < peers.size(); ii++)
 		{
 			if (!rbc->DeliverFrom(xtest, ii))
-				std::cout << "<X> ";
+				std::cerr << "<X> ";
 			else
-				std::cout << xtest << " ";
+				std::cerr << xtest << " ";
 		}
-		std::cout << std::endl;
+		std::cerr << std::endl;
 		mpz_clear(xtest);
 	}
 
@@ -214,11 +229,11 @@ void run_instance
 			dss_in << dss_c_ik[i][k] << std::endl;
 	}
 	if (opt_verbose)
-		std::cout << "CanettiGennaroJareckiKrawczykRabinDSS(in, ...)" << std::endl;
+		std::cerr << "INFO: CanettiGennaroJareckiKrawczykRabinDSS(in, ...)" << std::endl;
 	CanettiGennaroJareckiKrawczykRabinDSS *dss = new CanettiGennaroJareckiKrawczykRabinDSS(dss_in);
 	if (!dss->CheckGroup())
 	{
-		std::cerr << "R_" << whoami << ": " << "tDSS domain parameters are not correctly generated!" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": " << "tDSS domain parameters are not correctly generated!" << std::endl;
 		delete dss, delete rbc, delete aiou, delete aiou2;
 		release_mpis();
 		exit(-1);
@@ -226,53 +241,49 @@ void run_instance
 	// update the tDSS keys (proactive security against mobile adversary)
 	std::stringstream err_log;
 	if (opt_verbose)
-		std::cout << "R_" << whoami << ": dss.Refresh()" << std::endl;
-	if (!dss->Refresh(peers.size(), whoami, idx2dkg, dkg2idx, aiou, rbc, err_log, false))
+		std::cerr << "INFO: R_" << whoami << ": dss.Refresh()" << std::endl;
+	if (!dss->Refresh(peers.size(), whoami, idx2dkg, dkg2idx, aiou, rbc, err_log, false, cache, cache_mod, &cache_avail))
 	{
-		std::cerr << "R_" << whoami << ": " << "tDSS Refresh() failed" << std::endl;
-		std::cerr << "R_" << whoami << ": log follows " << std::endl << err_log.str();
+		std::cerr << "ERROR: R_" << whoami << ": " << "tDSS Refresh() failed" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": log follows " << std::endl << err_log.str();
 		delete dss, delete rbc, delete aiou, delete aiou2;
 		release_mpis();
 		exit(-1);
 	}
 	if (opt_verbose > 1)
-		std::cout << "R_" << whoami << ": log follows " << std::endl << err_log.str();
+		std::cerr << "INFO: R_" << whoami << ": log follows " << std::endl << err_log.str();
 	// at the end: deliver some more rounds for still waiting parties
 	time_t synctime = aiounicast::aio_timeout_very_long;
 	if (opt_verbose)
-		std::cout << "R_" << whoami << ": waiting approximately " << (synctime * (T_RBC + 1)) << " seconds for stalled parties" << std::endl;
+		std::cerr << "INFO: R_" << whoami << ": waiting approximately " << (synctime * (T_RBC + 1)) << " seconds for stalled parties" << std::endl;
 	rbc->Sync(synctime);
 	// create an OpenPGP DSA-based primary key using refreshed values from tDSS
 	gcry_mpi_t p, q, g, h, y, n, t, i, qualsize, x_rvss_qualsize, x_i, xprime_i;
 	std::vector<gcry_mpi_t> qual, x_rvss_qual;
 	std::vector< std::vector<gcry_mpi_t> > c_ik;
-	if (!mpz_get_gcry_mpi(&p, dss->p))
+	p = gcry_mpi_new(2048);
+	if (!mpz_get_gcry_mpi(p, dss->p))
 	{
-		std::cerr << "R_" << whoami << ": mpz_get_gcry_mpi() failed for p" << std::endl;
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
-		exit(-1);
-	}
-	if (!mpz_get_gcry_mpi(&q, dss->q))
-	{
-		std::cerr << "R_" << whoami << ": mpz_get_gcry_mpi() failed for q" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": mpz_get_gcry_mpi() failed for p" << std::endl;
 		gcry_mpi_release(p);
 		delete dss, delete rbc, delete aiou, delete aiou2;
 		release_mpis();
 		exit(-1);
 	}
-	if (!mpz_get_gcry_mpi(&g, dss->g))
+	q = gcry_mpi_new(2048);
+	if (!mpz_get_gcry_mpi(q, dss->q))
 	{
-		std::cerr << "R_" << whoami << ": mpz_get_gcry_mpi() failed for g" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": mpz_get_gcry_mpi() failed for q" << std::endl;
 		gcry_mpi_release(p);
 		gcry_mpi_release(q);
 		delete dss, delete rbc, delete aiou, delete aiou2;
 		release_mpis();
 		exit(-1);
 	}
-	if (!mpz_get_gcry_mpi(&h, dss->h))
+	g = gcry_mpi_new(2048);
+	if (!mpz_get_gcry_mpi(g, dss->g))
 	{
-		std::cerr << "R_" << whoami << ": mpz_get_gcry_mpi() failed for h" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": mpz_get_gcry_mpi() failed for g" << std::endl;
 		gcry_mpi_release(p);
 		gcry_mpi_release(q);
 		gcry_mpi_release(g);
@@ -280,9 +291,10 @@ void run_instance
 		release_mpis();
 		exit(-1);
 	}
-	if (!mpz_get_gcry_mpi(&y, dss->y))
+	h = gcry_mpi_new(2048);
+	if (!mpz_get_gcry_mpi(h, dss->h))
 	{
-		std::cerr << "R_" << whoami << ": mpz_get_gcry_mpi() failed for y" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": mpz_get_gcry_mpi() failed for h" << std::endl;
 		gcry_mpi_release(p);
 		gcry_mpi_release(q);
 		gcry_mpi_release(g);
@@ -291,9 +303,10 @@ void run_instance
 		release_mpis();
 		exit(-1);
 	}
-	if (!mpz_get_gcry_mpi(&x_i, dss->x_i))
+	y = gcry_mpi_new(2048);
+	if (!mpz_get_gcry_mpi(y, dss->y))
 	{
-		std::cerr << "R_" << whoami << ": mpz_get_gcry_mpi() failed for x_i" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": mpz_get_gcry_mpi() failed for y" << std::endl;
 		gcry_mpi_release(p);
 		gcry_mpi_release(q);
 		gcry_mpi_release(g);
@@ -303,9 +316,13 @@ void run_instance
 		release_mpis();
 		exit(-1);
 	}
-	if (!mpz_get_gcry_mpi(&xprime_i, dss->xprime_i))
+	if (libgcrypt_secmem)
+		x_i = gcry_mpi_snew(2048);
+	else
+		x_i = gcry_mpi_new(2048);
+	if (!mpz_get_gcry_mpi(x_i, dss->x_i))
 	{
-		std::cerr << "R_" << whoami << ": mpz_get_gcry_mpi() failed for xprime_i" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": mpz_get_gcry_mpi() failed for x_i" << std::endl;
 		gcry_mpi_release(p);
 		gcry_mpi_release(q);
 		gcry_mpi_release(g);
@@ -316,6 +333,24 @@ void run_instance
 		release_mpis();
 		exit(-1);
 	}
+	if (libgcrypt_secmem)
+		xprime_i = gcry_mpi_snew(2048);
+	else
+		xprime_i = gcry_mpi_new(2048);
+	if (!mpz_get_gcry_mpi(xprime_i, dss->xprime_i))
+	{
+		std::cerr << "ERROR: R_" << whoami << ": mpz_get_gcry_mpi() failed for xprime_i" << std::endl;
+		gcry_mpi_release(p);
+		gcry_mpi_release(q);
+		gcry_mpi_release(g);
+		gcry_mpi_release(h);
+		gcry_mpi_release(y);
+		gcry_mpi_release(x_i);
+		gcry_mpi_release(xprime_i);
+		delete dss, delete rbc, delete aiou, delete aiou2;
+		release_mpis();
+		exit(-1);
+	}
 	n = gcry_mpi_set_ui(NULL, dss->n);
 	t = gcry_mpi_set_ui(NULL, dss->t);
 	i = gcry_mpi_set_ui(NULL, dss->i);
@@ -337,9 +372,10 @@ void run_instance
 		for (size_t k = 0; k <= dss->t; k++)
 		{
 			gcry_mpi_t tmp;
-			if (!mpz_get_gcry_mpi(&tmp, dss->dkg->x_rvss->C_ik[j][k]))
+			tmp = gcry_mpi_new(2048);
+			if (!mpz_get_gcry_mpi(tmp, dss->dkg->x_rvss->C_ik[j][k]))
 			{
-				std::cerr << "R_" << whoami << ": mpz_get_gcry_mpi() failed for dss->dkg->x_rvss->C_ik[j][k]" << std::endl;
+				std::cerr << "ERROR: R_" << whoami << ": mpz_get_gcry_mpi() failed for dss->dkg->x_rvss->C_ik[j][k]" << std::endl;
 				gcry_mpi_release(p);
 				gcry_mpi_release(q);
 				gcry_mpi_release(g);
@@ -359,6 +395,7 @@ void run_instance
 				for (size_t jj = 0; jj < c_ik.size(); jj++)
 					for (size_t kk = 0; kk < c_ik[jj].size(); kk++)
 						gcry_mpi_release(c_ik[jj][kk]);
+				gcry_mpi_release(tmp);
 				delete dss, delete rbc, delete aiou, delete aiou2;
 				release_mpis();
 				exit(-1); 
@@ -390,23 +427,22 @@ void run_instance
 			gcry_mpi_release(c_ik[j][k]);
 	
 	// export updated private keys in OpenPGP armor format
-	tmcg_octets_t all;
+	tmcg_openpgp_octets_t all;
 	std::string armor;
 	std::stringstream secfilename;
 	secfilename << peers[whoami] << "_dkg-sec.asc";
-	armor = "", all.clear();
 	all.insert(all.end(), sec.begin(), sec.end());
 	all.insert(all.end(), uid.begin(), uid.end());
 	all.insert(all.end(), uidsig.begin(), uidsig.end());
 	all.insert(all.end(), ssb.begin(), ssb.end());
 	all.insert(all.end(), subsig.begin(), subsig.end());
-	CallasDonnerhackeFinneyShawThayerRFC4880::ArmorEncode(5, all, armor);
+	CallasDonnerhackeFinneyShawThayerRFC4880::ArmorEncode(TMCG_OPENPGP_ARMOR_PRIVATE_KEY_BLOCK, all, armor);
 	if (opt_verbose > 1)
 		std::cout << armor << std::endl;
 	std::ofstream secofs((secfilename.str()).c_str(), std::ofstream::out | std::ofstream::trunc);
 	if (!secofs.good())
 	{
-		std::cerr << "R_" << whoami << ": opening private key file failed" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": opening private key file failed" << std::endl;
 		delete dss, delete rbc, delete aiou, delete aiou2;
 		release_mpis();
 		exit(-1);
@@ -414,7 +450,7 @@ void run_instance
 	secofs << armor;
 	if (!secofs.good())
 	{
-		std::cerr << "R_" << whoami << ": writing private key file failed" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": writing private key file failed" << std::endl;
 		delete dss, delete rbc, delete aiou, delete aiou2;
 		release_mpis();
 		exit(-1);
@@ -430,13 +466,13 @@ void run_instance
 	// release handles (unicast channel)
 	uP_in.clear(), uP_out.clear(), uP_key.clear();
 	if (opt_verbose)
-		std::cout << "R_" << whoami << ": aiou.numRead = " << aiou->numRead <<
+		std::cerr << "INFO: R_" << whoami << ": aiou.numRead = " << aiou->numRead <<
 			" aiou.numWrite = " << aiou->numWrite << std::endl;
 
 	// release handles (broadcast channel)
 	bP_in.clear(), bP_out.clear(), bP_key.clear();
 	if (opt_verbose)
-		std::cout << "R_" << whoami << ": aiou2.numRead = " << aiou2->numRead <<
+		std::cerr << "INFO: R_" << whoami << ": aiou2.numRead = " << aiou2->numRead <<
 			" aiou2.numWrite = " << aiou2->numWrite << std::endl;
 
 	// release asynchronous unicast and broadcast
@@ -452,6 +488,7 @@ char *gnunet_opt_passwords = NULL;
 char *gnunet_opt_port = NULL;
 unsigned int gnunet_opt_xtests = 0;
 unsigned int gnunet_opt_wait = 5;
+unsigned int gnunet_opt_W = opt_W;
 int gnunet_opt_verbose = 0;
 #endif
 
@@ -459,7 +496,7 @@ void fork_instance
 	(const size_t whoami)
 {
 	if ((pid[whoami] = fork()) < 0)
-		perror("dkg-refresh (fork)");
+		perror("ERROR: dkg-refresh (fork)");
 	else
 	{
 		if (pid[whoami] == 0)
@@ -471,14 +508,14 @@ void fork_instance
 			run_instance(whoami, 0);
 #endif
 			if (opt_verbose)
-				std::cout << "R_" << whoami << ": exit(0)" << std::endl;
+				std::cerr << "INFO: R_" << whoami << ": exit(0)" << std::endl;
 			exit(0);
 			/* END child code: participant R_i */
 		}
 		else
 		{
 			if (opt_verbose)
-				std::cout << "fork() = " << pid[whoami] << std::endl;
+				std::cerr << "INFO: fork() = " << pid[whoami] << std::endl;
 			instance_forked = true;
 		}
 	}
@@ -527,6 +564,12 @@ int main
 			"minutes to wait until start of refresh protocol",
 			&gnunet_opt_wait
 		),
+		GNUNET_GETOPT_option_uint('W',
+			"aiou-timeout",
+			"TIME",
+			"timeout for point-to-point messages in minutes",
+			&gnunet_opt_W
+		),
 		GNUNET_GETOPT_option_uint('x',
 			"x-tests",
 			NULL,
@@ -553,6 +596,8 @@ int main
 		passwords = gnunet_opt_passwords; // get passwords from GNUnet options
 	if (gnunet_opt_hostname != NULL)
 		hostname = gnunet_opt_hostname; // get hostname from GNUnet options
+	if (gnunet_opt_W != opt_W)
+		opt_W = gnunet_opt_W; // get aiou message timeout from GNUnet options
 #endif
 
 	// create peer list from remaining arguments
@@ -561,7 +606,7 @@ int main
 		std::string arg = argv[i+1];
 		// ignore options
 		if ((arg.find("-c") == 0) || (arg.find("-p") == 0) || (arg.find("-r") == 0) || (arg.find("-w") == 0) || (arg.find("-L") == 0) || 
-			(arg.find("-l") == 0) || (arg.find("-x") == 0) || (arg.find("-P") == 0) || (arg.find("-H") == 0))
+			(arg.find("-l") == 0) || (arg.find("-x") == 0) || (arg.find("-P") == 0) || (arg.find("-H") == 0) || (arg.find("-W") == 0))
 		{
 			size_t idx = ++i;
 			if ((arg.find("-H") == 0) && (idx < (size_t)(argc - 1)) && (opt_hostname == NULL))
@@ -576,6 +621,8 @@ int main
 			}
 			if ((arg.find("-p") == 0) && (idx < (size_t)(argc - 1)) && (port.length() == 0))
 				port = argv[i+1];
+			if ((arg.find("-W") == 0) && (idx < (size_t)(argc - 1)) && (opt_W == 5))
+				opt_W = strtoul(argv[i+1], NULL, 10);
 			continue;
 		}
 		else if ((arg.find("--") == 0) || (arg.find("-v") == 0) || (arg.find("-h") == 0) || (arg.find("-V") == 0))
@@ -592,6 +639,7 @@ int main
 				std::cout << "  -P STRING      exchanged passwords to protect private and broadcast channels" << std::endl;
 				std::cout << "  -v, --version  print the version number" << std::endl;
 				std::cout << "  -V, --verbose  turn on verbose output" << std::endl;
+				std::cout << "  -W TIME        timeout for point-to-point messages in minutes" << std::endl;
 #endif
 				return 0; // not continue
 			}
@@ -628,11 +676,19 @@ int main
 	peers.push_back("Test3");
 	opt_verbose = 1;
 #endif
+
+	// check command line arguments
+	if ((opt_hostname != NULL) && (opt_passwords == NULL))
+	{
+		std::cerr << "ERROR: option \"-P\" is necessary due to insecure network" << std::endl;
+		return -1;
+	}
 	if (peers.size() < 1)
 	{
 		std::cerr << "ERROR: no peers given as argument; usage: " << usage << std::endl;
 		return -1;
 	}
+
 	// canonicalize peer list
 	std::sort(peers.begin(), peers.end());
 	std::vector<std::string>::iterator it = std::unique(peers.begin(), peers.end());
@@ -642,22 +698,97 @@ int main
 		std::cerr << "ERROR: too few or too many peers given" << std::endl;
 		return -1;
 	}
+	if (opt_verbose)
+	{
+		std::cerr << "INFO: canonicalized peer list = " << std::endl;
+		for (size_t i = 0; i < peers.size(); i++)
+			std::cerr << peers[i] << std::endl;
+	}
+
+	// lock memory
+	if (!lock_memory())
+	{
+		std::cerr << "WARNING: locking memory failed; CAP_IPC_LOCK required for memory protection" << std::endl;
+		// at least try to use libgcrypt's secure memory
+		if (!gcry_check_version(TMCG_LIBGCRYPT_VERSION))
+		{
+			std::cerr << "ERROR: libgcrypt version >= " << TMCG_LIBGCRYPT_VERSION << " required" << std::endl;
+			return -1;
+		}
+		gcry_control(GCRYCTL_SUSPEND_SECMEM_WARN);
+		gcry_control(GCRYCTL_USE_SECURE_RNDPOOL);
+		gcry_control(GCRYCTL_INIT_SECMEM, 16384, 0);
+		gcry_control(GCRYCTL_RESUME_SECMEM_WARN);
+		gcry_control(GCRYCTL_INITIALIZATION_FINISHED, 0);
+		libgcrypt_secmem = true;
+	}
+
+	// initialize LibTMCG
 	if (!init_libTMCG())
 	{
 		std::cerr << "ERROR: initialization of LibTMCG failed" << std::endl;
 		return -1;
 	}
-	if ((opt_hostname != NULL) && (opt_passwords == NULL))
+	if (opt_verbose)
+		std::cerr << "INFO: using LibTMCG version " << version_libTMCG() << std::endl;
+
+	// read the public key
+	std::string armored_pubkey = "undefined";
+	if (opt_hostname != NULL)
 	{
-		std::cerr << "ERROR: option \"-P\" is necessary due to insecure network" << std::endl;
+		if (!read_key_file(hostname + "_dkg-pub.asc", armored_pubkey))
+			armored_pubkey = "undefined";
+	}
+	else
+	{
+		for (size_t i = 0; i < peers.size(); i++)
+		{
+			if (read_key_file(peers[i] + "_dkg-pub.asc", armored_pubkey))
+				break;
+		}
+	}
+	if (armored_pubkey == "undefined")
+	{
+		std::cerr << "ERROR: no corresponding public key file found" << std::endl;
 		return -1;
 	}
-	if (opt_verbose)
+
+	// parse the public key block and corresponding signatures
+	TMCG_OpenPGP_Pubkey *primary = NULL;
+	bool parse_ok = CallasDonnerhackeFinneyShawThayerRFC4880::
+		PublicKeyBlockParse(armored_pubkey, opt_verbose, primary);
+	if (parse_ok)
 	{
-		std::cout << "INFO: canonicalized peer list = " << std::endl;
-		for (size_t i = 0; i < peers.size(); i++)
-			std::cout << peers[i] << std::endl;
+		if (primary->pkalgo != TMCG_OPENPGP_PKALGO_DSA)
+		{
+			std::cerr << "ERROR: primary key is not DSA" << std::endl;
+			delete primary;
+			return -1;
+		}
 	}
+	else
+	{
+		std::cerr << "ERROR: cannot use the provided public key" << std::endl;
+		if (primary)
+			delete primary;
+		return -1;
+	}
+
+	// initialize cache
+	mpz_init(dss_q);
+	if (!mpz_set_gcry_mpi(primary->dsa_q, dss_q))
+	{
+		std::cerr << "ERROR: mpz_set_gcry_mpi() failed for dss_q" << std::endl;
+		release_mpis();
+		return -1;
+	}
+	delete primary;
+	std::cerr << "We need some entropy to cache very strong randomness for share refresh." << std::endl;
+	std::cerr << "Please use other programs, move the mouse, and type on your keyboard: " << std::endl; 
+	mpz_ssrandomm_cache_init(cache, cache_mod, &cache_avail, (2 * peers.size()), dss_q);
+	std::cerr << "Thank you!" << std::endl;
+	mpz_clear(dss_q);
+
 	// initialize return code
 	int ret = 0;
 	// create underlying point-to-point channels, if built-in TCP/IP service requested
@@ -677,6 +808,8 @@ int main
 		ret = tcpip_io();
 		tcpip_close();
 		tcpip_done();
+		// release cache
+		mpz_ssrandomm_cache_done(cache, cache_mod, &cache_avail);
 		return ret;
 	}
 
@@ -712,6 +845,12 @@ int main
 			"minutes to wait until start of refresh protocol",
 			&gnunet_opt_wait
 		),
+		GNUNET_GETOPT_option_uint('W',
+			"aiou-timeout",
+			"TIME",
+			"timeout for point-to-point messages in minutes",
+			&gnunet_opt_W
+		),
 		GNUNET_GETOPT_option_uint('x',
 			"x-tests",
 			NULL,
@@ -722,24 +861,27 @@ int main
 	};
 	ret = GNUNET_PROGRAM_run(argc, argv, usage, about, myoptions, &gnunet_run, argv[0]);
 	GNUNET_free((void *) argv);
+	// release cache
+	mpz_ssrandomm_cache_done(cache, cache_mod, &cache_avail);
 	if (ret == GNUNET_OK)
 		return 0;
 	else
 		return -1;
 #else
-	std::cerr << "WARNING: GNUnet development files are required for message exchange of refresh protocol" << std::endl;
+	std::cerr << "WARNING: GNUnet CADET is required for the message exchange of this program" << std::endl;
 #endif
 
-	std::cout << "INFO: running local test with " << peers.size() << " participants" << std::endl;
+	std::cerr << "INFO: running local test with " << peers.size() << " participants" << std::endl;
+	std::cerr << "WARNING: due to cache issues the generated shares are identical, don't use them!" << std::endl;
 	// open pipes
 	for (size_t i = 0; i < peers.size(); i++)
 	{
 		for (size_t j = 0; j < peers.size(); j++)
 		{
 			if (pipe(pipefd[i][j]) < 0)
-				perror("dkg-refresh (pipe)");
+				perror("ERROR: dkg-refresh (pipe)");
 			if (pipe(broadcast_pipefd[i][j]) < 0)
-				perror("dkg-refresh (pipe)");
+				perror("ERROR: dkg-refresh (pipe)");
 		}
 	}
 	
@@ -755,9 +897,9 @@ int main
 	{
 		int wstatus = 0;
 		if (opt_verbose)
-			std::cout << "waitpid(" << pid[i] << ")" << std::endl;
+			std::cerr << "INFO: waitpid(" << pid[i] << ")" << std::endl;
 		if (waitpid(pid[i], &wstatus, 0) != pid[i])
-			perror("dkg-refresh (waitpid)");
+			perror("ERROR: dkg-refresh (waitpid)");
 		if (!WIFEXITED(wstatus))
 		{
 			std::cerr << "ERROR: protocol instance ";
@@ -770,19 +912,22 @@ int main
 		else if (WIFEXITED(wstatus))
 		{
 			if (opt_verbose)
-				std::cout << "INFO: protocol instance " << pid[i] << " terminated with exit status " << WEXITSTATUS(wstatus) << std::endl;
+				std::cerr << "INFO: protocol instance " << pid[i] << " terminated with exit status " << WEXITSTATUS(wstatus) << std::endl;
 			if (WEXITSTATUS(wstatus))
 				ret = -2; // error
 		}
 		for (size_t j = 0; j < peers.size(); j++)
 		{
 			if ((close(pipefd[i][j][0]) < 0) || (close(pipefd[i][j][1]) < 0))
-				perror("dkg-refresh (close)");
+				perror("ERROR: dkg-refresh (close)");
 			if ((close(broadcast_pipefd[i][j][0]) < 0) || (close(broadcast_pipefd[i][j][1]) < 0))
-				perror("dkg-refresh (close)");
+				perror("ERROR: dkg-refresh (close)");
 		}
 	}
 	
+	// release cache
+	mpz_ssrandomm_cache_done(cache, cache_mod, &cache_avail);
+	// finish	
 	return ret;
 }
 
